/* Generate 2-designs.
   Version 1.7, June 2023.  Brendan McKay.
   Subject to the same licence as nauty 2.8. */

#include "gtools.h"
#include "naumdesign.h"
#ifdef ZLIB
#include <zlib.h>
#endif

#ifdef PROB
#define pUSE " [-p#/#]..."
#define pHELP \
   "     -p#/#  -pN/L means expand nodes at level L with probability 1/N.\n"
#else
#define pUSE ""
#define pHELP  ""
#endif

#ifdef ZLIB
#define ZLIBHELP \
"     -z   gzip output file\n"
#define ZLIBUSE "|-z"
#else
#define ZLIBHELP ""
#define ZLIBUSE ""
#endif

#if WORDSIZE==64
#if PROB
#define PROGNAME "gen2des_p"
#else
#define PROGNAME "gen2des64"
#endif
#else
#define PROGNAME "gen2des"
#endif

#define USAGE \
  PROGNAME " [-q] [-P#] [-s|-i#] [-a#] [-o#|-S] [-F] [-L] [-u" \
     ZLIBUSE "] [-v] [#/#/#] v k lambda" pUSE

#define HELPTEXT \
"  Generate 2-designs with parameters 2-(v,k,lambda)\n" \
"\n" \
"     -s   Only generate simple designs\n" \
"     -i#  Blocks can intersect in at most this many points\n" \
"     -a#  Only output designs with at least # automorphisms\n" \
"     -o#  Specify one compulsory block overlap\n" \
"     -S   Special processing for Steiner designs (lambda=1)\n" \
"     -Q   Don't generate, just report on the parameters\n" \
"     -L   Only generate up to the splitting level and report\n" \
"     -F   Flush output after each design is found\n" \
"     -u   No output, just count\n" \
ZLIBHELP \
"     -P#  Write a progress message after each 1000*# nodes\n" \
"     -v   More verbose output\n" \
"     -q   Suppress auxiliary output.\n" \
"\n" \
"     #/#/#  r/m/l means only generate part r out of parts 0..m-1\n" \
"          defined at level l.\n" \
pHELP

#if MULTBITS <= 8
typedef unsigned char multint;
#else
typedef unsigned short multint;
#endif

#ifndef INVAR
#define INVAR invar2b
#endif

static int v,k,lambda,b,r;
static boolean simple,uswitch,flush,Steiner;
static unsigned long long minautoms;
static int mininter,maxinter,maxmult;
static setword nointmask,yesintmask;
static int eqmultmax;
static int splitlevel,splitcount,splitmod;
static boolean splittest;
static long long casecounter;
static int olap;

static struct stage
{
    design des[MAXBLOCKS];
    size_t extsize;  /* size of ext[] in multints */
    size_t numext;   /* number of records in ext[] */
    multint *ext;    /* List of all extensions */
    size_t eorbsize; /* size of borb[] in longs */
    long *eorb;      /* Orbits of extensions */
    long invar[MAXPOINTS+1];
} stage[MAXPOINTS+1];

/* Automorphisms in action on blocks stored in rows: */
static int autom[MAXPOINTS][MAXBLOCKS];
static int numautom;

static long long numwritten,nout,nsimple;
static long long nodecount[MAXPOINTS],nontriv[MAXPOINTS];
static long long refcalls,nautycalls,reffails;
static long long maxnumext,totnodes,progress;

typedef multint* multptr;
static int mbval;
static multptr *extptr;
static size_t extptrsize;

static int minfull[MAXPOINTS+1];

static unsigned char line[20+MAXBLOCKS*(1+(MAXPOINTS+3)/4)];
static unsigned char *linepos;

static unsigned long long groupsize;  /* Only calculated at output size */
static boolean transitive;            /* Only used at output size */
#define MAXGROUPS 200   /* Maximum number of different group sizes */
static struct
{
   unsigned long long groupsize;
   long long totalnum,simplenum;
   long long trans,simpletrans;
} groupcount[MAXGROUPS];
static int numgroups;

#ifdef PROB
static long prob[MAXPOINTS+1];
static unsigned long long avail[MAXPOINTS+1],used[MAXPOINTS+1];
#endif

#define SORT_OF_SORT 4
#define SORT_NAME sortexts
#define SORT_TYPE1 multptr
#define SORT_COMPARE(x,y) extcompare(x,y,mbval)

static int
extcompare(multint *x, multint *y, int mb)
{
    int i;

    for (i = 0; i < mb; ++i) if (x[i] != y[i]) break;

    if (i == mb) return 0;
    else if (x[i] < y[i]) return -1;
    else                  return 1;
}

#include "sorttemplates.c"

#define SORT_OF_SORT 3
#define SORT_NAME sortindirect
#define SORT_TYPE1 int
#define SORT_TYPE2 long
#include "sorttemplates.c"

#define ACCUM(x,y)   x = (((x) + (y)) & 077777)

static int fuzz1[] = {037541,061532,005257,026416};
static int fuzz2[] = {006532,070236,035523,062437};

#define FUZZ1(x) ((x) ^ fuzz1[(x)&3])
#define FUZZ2(x) ((x) ^ fuzz2[(x)&3])
#define MIN(x,y) ((x)<(y) ? (x) : (y))
#define MAX(x,y) ((x)>(y) ? (x) : (y))

#define CHECKERROR(id,msg) \
{ dummy(id,s,msg); fprintf(stderr,">E id=%d s=%d\n",id,s); gt_abort(msg); }

/*******************************************************************/

static unsigned long long
deshash(design *des, int s)
/* Hash code for design, intended for debugging */
{
    int q,mb;
    unsigned long long ans;

    mb = des[0];
    ans = s;

    for (q = 1; q <= mb; ++q)
        ans = 73 * ans + des[q];

    return ans;
}

/*******************************************************************/

#if 0
static void
writestage(int s, int detail)
{
    int i,mb,q;
    design *des;
    multint *ext;

    des = stage[s].des;
    mb = des[0];
    printf("------\nStage %d [%lld]; des=%llu; mb=%d; numext=%lu\n",
            s,nodecount[s],deshash(stage[s].des,s),mb,stage[s].numext);
    if (detail == 1) return;

    printf("des=");
    for (q = 1; q <= mb; ++q)
        printf(" %2lx*%lu",
            (unsigned long)(BLOCK(des[q])>>(WORDSIZE-MULTBITS)),
            (unsigned long)MULT(des[q]));
    printf("\n");
    if (detail == 2) return;

    ext = stage[s].ext;
    printf("ext=\n");
    for (q = 0; q < stage[s].numext; ++q)
    {
        for (i = 0; i < mb; ++i) printf(" %2d",ext[q*mb+i]);
        printf("\n");
    }
    printf("automs=\n");
    for (q = 0; q < numautom; ++q)
    {
        for (i = 0; i < mb; ++i) printf(" %2d",autom[q][i]);
        printf("\n");
    }
}
#endif

/*******************************************************************/

static void
putdesign(FILE *f, int s, boolean next)
{
    design *des;
    int q,i;
    setword w;

    des = stage[s].des;
    fprintf(f,"[%d,%lu] ",s,(long unsigned)des[0]);
    for (q = 1; q <= des[0]; ++q)
    {
        w = BLOCK(des[q]);
        if (w == 0) fprintf(f,"zero");
        while (w)
        {
            TAKEBIT(i,w);
            if (w == 0) fprintf(f,"%d",i);
            else        fprintf(f,"%d,",i);
        }
        if (MULT(des[q]) > 1) fprintf(f,"*%d ",(int)MULT(des[q]));
        else                  fprintf(f," ");
    }
    if (next) fprintf(f," ext=%ld\n",(long)stage[s].numext);
    else      fprintf(f,"\n");
}

static void
dummy(int id, int s, char *msg)
{
    printf(" ");
}

#if 0
static void
check_stage(int id, int s, boolean checkgroup, boolean checkext)
/* Check stage s for sanity, optionally including the extension list. */
{
    int mb,totmult,tot,i,j,q,qq;
    design *des;
    setword w;

    des = stage[s].des;
    mb = des[0];

    /* Check blocks are in order and the total multiplicity is correct */

    totmult = 0;
    for (q = 1; q <= mb; ++q)
    {
        totmult += MULT(des[q]);
        if (q > 1 && BLOCK(des[q]) <= BLOCK(des[q-1]))
            CHECKERROR(id,">E check: blocks not in order\n");
    }

    if (totmult != b)
        CHECKERROR(id,">E check: total multiplicity is wrong\n");
  
    /* Check point coverage */

    for (i = 0; i < s; ++i)
    {
        tot = 0;
        w = bit[i];
        for (q = 1; q <= mb; ++q)
            if ((des[q]&w) == w) tot += MULT(des[q]);
        if (tot != r) CHECKERROR(id,">E check: point coverage is wrong\n");
    } 
  
    /* Check pair coverage */

    for (i = 0; i < s; ++i)
    for (j = i+1; j < s; ++j)
    {
        tot = 0;
        w = bit[i] | bit[j];
        for (q = 1; q <= mb; ++q)
            if ((des[q]&w) == w) tot += MULT(des[q]);
        if (tot != lambda) CHECKERROR(id,">E check: pair coverage is wrong\n");
    } 

    /* Check block sizes */

    for (q = 1; q <= mb; ++q)
    {
        tot = BLOCKLEN(des[q]);
        if (tot > k) CHECKERROR(id,">E check: block too large\n");
        if (tot+v-s < k) CHECKERROR(id,">E check: block too small\n");
    }

    /* Check multiplicity bound */

    for (q = 1; q <= mb; ++q)
    {
        if (BLOCKLEN(des[q]) == k && MULT(des[q]) > maxmult)
            CHECKERROR(id,">E multiplicity too large (1)\n");

        if (BLOCKLEN(des[q])+v-s == k && MULT(des[q]) > maxmult)
            CHECKERROR(id,">E multiplicity too large (2)\n");
    }

    /* Check intersection in symmetric case */

    if (v == b)
    {
        for (q = 1; q < mb; ++q)   /* < is correct */
        for (qq = q+1; qq <= mb; ++qq)
        {
            tot = POPCOUNT(BLOCK(des[q])&BLOCK(des[qq]));
            if (BLOCKLEN(des[q]) == k || BLOCKLEN(des[qq]) == k)
            {
                if (tot != lambda)
                    CHECKERROR(id,">E symmetric wrong intersection\n");
            }
            else if (tot > lambda)
                CHECKERROR(id,">E symmetric excessive intersection\n");
        }
    }
}
#endif

/*******************************************************************/

static void
addgroup(boolean issimple)
/* Add groupsize to table; groupsize and transitive are global */ 
{
    int i,j;

    for (i = 0; i < numgroups && groupcount[i].groupsize < groupsize; ++i)
        { }

    if (i < numgroups && groupcount[i].groupsize == groupsize)
    {
        ++groupcount[i].totalnum;
        if (issimple) ++groupcount[i].simplenum;
        if (transitive) 
        {
            ++groupcount[i].trans;
            if (issimple) ++groupcount[i].simpletrans;
        }
        return;
    }

    if (numgroups == MAXGROUPS)
        gt_abort(">E gen2des : increase MAXGROUPS\n");

    for (j = numgroups; j > i; --j)
        groupcount[j] = groupcount[j-1];
    groupcount[i].groupsize = groupsize;
    groupcount[i].totalnum = 1;
    groupcount[i].simplenum = (issimple ? 1 : 0);
    if (transitive)
    {
        groupcount[i].trans = 1;
        groupcount[i].simpletrans = (issimple ? 1 : 0);
    }
    else
        groupcount[i].trans = groupcount[i].simpletrans = 0;
    ++numgroups;
}

#ifdef ZLIB
static gzFile gzfile;
#endif
static boolean gzip;

static void
got_one()
/* Called when stage[v] has an accepted full design */
{
    int i,j,q,mb,i0,m;
    design *des;
    static unsigned char hex[]
       = {'0','1','2','3','4','5','6','7','8','9','a','b','c','d','e','f'};
    unsigned char *s;
    setword w;

    ++nout;

    des = stage[v].des;
    mb = des[0];
    if (mb == b) { ++nsimple; addgroup(TRUE); }
    else         { addgroup(FALSE); }    

    if (uswitch) return;

    if (groupsize < minautoms && groupsize != 0) return;

    ++numwritten;

    s = linepos;
    i0 = 4 * ((v+3)/4 - 1);
    for (q = 1; q <= mb; ++q)
    {
        w = des[q] >> (WORDSIZE - v);
        m = MULT(des[q]);
        for (j = 0; j < m; ++j)
        {
            *(s++) = ' ';
            for (i = i0; i >= 0; i -= 4) *(s++) = hex[(w>>i)&0xf];
        }
    }
    *(s++) = '\n';

#ifdef ZLIB
    if (gzip)
    {
        gzwrite(gzfile,line,s-line);
    }
    else
#endif
    {
        fwrite(line,1,s-line,stdout);
        if (flush) fflush(stdout);
    }
}

/*******************************************************************/

static void
invar1(int s)
/* Calculate the invariant for stage s from that for stage s-1 */
{
    int i,j,q,mb,mult;
    long *inv0,*inv1;
    design *des;
    setword x,y,bitsm,mask;

    inv0 = stage[s-1].invar;
    inv1 = stage[s].invar;

    des = stage[s].des;
    mb = des[0];

    if (s >= v-1)
    {
        for (i = 0; i < s; ++i) inv1[i] = 0;

        for (j = 1; j < s; ++j)
        for (i = 0; i < j; ++i)
        {
            x = bit[i] | bit[j];
            for (q = 1; q <= mb; ++q)
            {
                if ((x & des[q]) == x && MULT(des[q]) > 1)
                {
                    ++inv1[i];
                    ++inv1[j];
                }
            } 
        }
    }
    else
    {
        for (i = 0; i < s-1; ++i) inv1[i] = inv0[i];
        inv1[s-1] = 0;
        bitsm = bit[s-1];
        mask = BLOCKMASK & ~bitsm;

        for (q = 1; q <= mb; ++q)
        {
            if ((des[q] & bitsm))
            {
                mult = MULT(des[q]);
                x = des[q] & mask;
                while (x)
                {
                    TAKEBIT(i,x);
                y = x;
                while (y)
                {
                    TAKEBIT(j,y);
                    inv1[i] += mult;
                    inv1[j] += mult;
                    inv1[s-1] += mult;
                }
                }
            }
        }
    }
}

static void
invar2(int s)
/* Calculate the invariant for stage s from that for stage s-1 */
{
    int i,j,q,mb,mbs;
    long *inv0,*inv1;
    design *des;
    setword mask;
    long count,extra;
    design sdes[MAXBLOCKS];

    inv0 = stage[s-1].invar;
    inv1 = stage[s].invar;

    des = stage[s].des;
    mb = des[0];

    for (i = 0; i < s-1; ++i) inv1[i] = inv0[i];
    inv1[s-1] = 0;

    mask = bit[s-1];
    mbs = 0;
    for (q = 1; q <= mb; ++q)
        if ((des[q] & mask))
        {
            sdes[mbs] = des[q];
            ++mbs;
        }

    for (i = 0; i < s-2; ++i)
    for (j = i+1; j < s-1; ++j)
    {
        count = 0;
        mask = bit[i] | bit[j];
        for (q = 0; q < mbs; ++q)
            if ((sdes[q] & mask) == mask) count += MULT(sdes[q]); 
        count ^= 0x3455; 
        for (q = 0; q < mbs; ++q)
            if ((sdes[q] & mask) == mask)
            {
                extra = count * MULT(sdes[q]);
                inv1[i] += extra;
                inv1[j] += extra;
                inv1[s-1] += extra;
            }
    }
}


static void
invar2b(int s)
/* Calculate the invariant for stage s from that for stage s-1 */
{
    int i,j,q,mb,mbs;
    long *inv0,*inv1;
    design *des;
    setword mask;
    long count,extra;
    design sdes[MAXBLOCKS];

    inv0 = stage[s-1].invar;
    inv1 = stage[s].invar;

    des = stage[s].des;
    mb = des[0];

    for (i = 0; i < s-1; ++i) inv1[i] = inv0[i];
    inv1[s-1] = 0;

    mask = bit[s-1];
    mbs = 0;
    for (q = 1; q <= mb; ++q)
        if ((des[q] & mask))
        {
            sdes[mbs] = des[q];
            ++mbs;
        }

    for (i = 0; i < s-2; ++i)
    for (j = i+1; j < s-1; ++j)
    {
        count = 0;
        mask = bit[i] | bit[j];
        for (q = 0; q < mbs; ++q)
            if ((sdes[q] & mask) == mask) count += MULT(sdes[q]); 
        extra = (17*count) ^ 2761;
        inv1[i] += extra;
        inv1[j] += extra;
        inv1[s-1] += extra;
    }
}

static void
invar3(int s)
/* Calculate the invariant for stage s from that for stage s-1 */
/* Only value for k=3 !!  */
{
    int i,j,q,mb,mult;
    long *inv0,*inv1;
    design *des;
    setword bitsm,x;

    if (k != 3) gt_abort(">E gen2des: updateinv3 is only valid for k=3\n");

    inv0 = stage[s-1].invar;
    inv1 = stage[s].invar;

    des = stage[s].des;
    mb = des[0];

    for (i = 0; i < s-1; ++i) inv1[i] = inv0[i];
    inv1[s-1] = 0;
    bitsm = bit[s-1];

    for (q = 1; q <= mb; ++q)
    {
        if ((des[q] & bitsm) && BLOCKLEN(des[q]) == 3)
        {
            x = des[q];
            mult = MULT(des[q]);
            TAKEBIT(j,x);
            inv1[j] += mult;
            TAKEBIT(j,x);
            inv1[j] += mult;
            inv1[s-1] += mult;
        }
    }
}

/*******************************************************************/

static void
moreext(int s)
/* More extension space for stage s */
{
    size_t len;

    len = 2*stage[s].extsize;
    if ((stage[s].ext = realloc(stage[s].ext,len*sizeof(multint))) == NULL)
        gt_abort(">E gen2des: can't extend ext\n");
    stage[s].extsize = len;
}

/*******************************************************************/

static void
automproc(int count, int *p, int *orb, int numorb, int stabv, int n)
/* Collect automorphisms converted into block action */
{
    blockautom(stage[n].des,p,autom[count-1],n);
    numautom = count;

#if 0
{
    int i,q,mb;
    design *des;

    des = stage[n].des;
    mb = des[0];

    printf("autom %d. des=",count);
    for (q = 1; q <= mb; ++q)
        printf(" %2lx*%lu",BLOCK(des[q])>>(WORDSIZE-MULTBITS),MULT(des[q]));
    printf("\np = ");
    for (i = 0; i < n; ++i) printf(" %2d",p[i]);
    printf("\nbp = ");
    for (i = 0; i < mb; ++i) printf(" %2d",autom[count-1][i]);
    printf("\n");
}
#endif
}

/*******************************************************************/

static void
makeeorbits(int s)
/* Make extension orbits for stage s, non-trivial group only */
{
    size_t numext,i,mb;
    ssize_t bot,mid,top;
    size_t q,qp,j1,j2;
    multint *ext,*extmid,newext[MAXBLOCKS];
    int *p,ip;
    long *eorb;

    numext = stage[s].numext;
    if (stage[s].eorbsize < numext)
    {
        if (stage[s].eorb) free(stage[s].eorb);
        if ((stage[s].eorb
               = malloc((numext+100)*sizeof(long))) == NULL)
            gt_abort(">E gen2des: can't extend eorb\n");
        stage[s].eorbsize = numext + 100;
    }
    if (extptrsize < numext)
    {
        if (extptr) free(extptr);
        if ((extptr = malloc((numext+100)*sizeof(multint*))) == NULL)
            gt_abort(">E gen2des: can't extend eorb\n");
        extptrsize = numext + 100;
    }

    mb = stage[s].des[0];
    mbval = mb;
    ext = stage[s].ext;
    eorb = stage[s].eorb;

    for (i = 0; i < numext; ++i) extptr[i] = ext + mb*i;
    sortexts(extptr,numext);

    for (i = 0; i < numext; ++i) eorb[i] = i;

    for (ip = 0; ip < numautom; ++ip)
    {
        p = autom[ip];

        for (q = 0; q < numext; ++q)
        {
            for (i = 0; i < mb; ++i)
                newext[i] = ext[q*mb+p[i]];
            bot = 0;
            top = numext-1;
            while (bot <= top)
            {
                mid = bot + (top-bot)/2;
                extmid = extptr[mid];
                for (i = 0; i < mb; ++i)
                    if (extmid[i] != newext[i]) break;
                if (i == mb)
                    break;
                else if (extmid[i] < newext[i])
                    bot = mid + 1;
                else
                    top = mid - 1;
            }
            qp = (extmid - ext) / mb;

            if (bot > top)
            {
                fprintf(stderr,"Binary search fail ip=%d/%d, q=%d/%d\n",
                        ip,numautom,(int)q,(int)numext);
                gt_abort(">E gen2des: binary search fail\n");
            }

            j1 = eorb[q];
            while (eorb[j1] != j1) j1 = eorb[j1];
            j2 = eorb[qp];
            while (eorb[j2] != j2) j2 = eorb[j2];

            if      (j1 < j2) eorb[j2] = eorb[q] = eorb[qp] = j1;
            else if (j1 > j2) eorb[j1] = eorb[q] = eorb[qp] = j2;
        }
    }
}

/*******************************************************************/

static long
pointinvar(int s, int x, long *invar)
/* Make an invariant value for point x that depends on invar */
{
    int i,j,q,mb;
    long ans,extra;
    design *des;
    setword mask;

    des = stage[s].des;
    mb = des[0];

    ans = invar[x];
    for (i = 0; i < s-1; ++i)
    if (i != x)
        for (j = i+1; j < s; ++j)
        if (j != x)
        {
            mask = bit[i] | bit[x] | bit[j];
            extra = invar[i] ^ invar[j];
            for (q = 1; q <= mb; ++q)
                if ((des[q] & mask) == mask)
                    ans += extra * MULT(des[q]);
        }

    return ans;
}

/*******************************************************************/

static void
makegroupsize(int *lab, int *ptn, int level, int *orbits, statsblk *stats,
              int tv, int index, int tcellsize, int numcells, int cc, int n)
/* Accumulate group size; values >= 2^64 are set to 0 */
{
    unsigned long long oldgroupsize;

    oldgroupsize = groupsize;
    groupsize = oldgroupsize * (unsigned long long)index;
    if (groupsize / (unsigned long long)index != oldgroupsize)
        groupsize = 0;
}

/*******************************************************************/

static boolean
pathinvar(design *des, int s, int *inv)
/* Compute a point invariant that should be minimal at point s-1.
   If point s-1 is not minimal, return FALSE without computing all
   the invariant values. Otherwise TRUE. */
{
    int mb,q,q1,q2,max,i,j,val,nfb;
    setword bitj,all,good,full[MAXBLOCKS];

    mb = des[0];
    max = (s > k ? k : s);

    if (olap >= 0 && s > 2*k-olap)
    {
       /* Misses of two full blocks overlapping in olap points */

        all = ALLMASK(s);
        nfb = 0;
        for (q = 1; q <= mb; ++q)
            if (BLOCKLEN(des[q]) == k)
                 full[nfb++] = BLOCK(des[q]);

        good = 0;
        for (q2 = 1; q2 < nfb; ++q2)
        for (q1 = 0; q1 < q2; ++q1)
            if (POPCOUNT(full[q1]&full[q2]) == olap)
                good |= all & ~(full[q1] | full[q2]);

        if (!(good & bit[s-1])) return FALSE;
        for (i = 0; i < s; ++i)
            if ((good & bit[i])) inv[i] = 0;
            else                 inv[i] = 1;
        return TRUE;
    }
    else
    {
       /* Number of blocks of size max */

        for (j = s; --j >= 0;)
        {
            bitj = bit[j];
            val = 0;
            for (q = 1; q <= mb; ++q)
                if ((des[q]&bitj) && BLOCKLEN(des[q]) == max)
                    val += MULT(des[q]);
            if (j < s-1 && val < inv[s-1]) return FALSE;
            inv[j] = val;
        }
    }

    return TRUE;
}

/* Special code for Steiner systems (lambda=1) with -S.

   Consider a set of partial blocks, all of them of length at least
   two, at most one of them not complete, whose intersection is a
   single point (the hub). This is called a flower if it covers all
   the points of the partial design. An active point of a flower is:
   -- if there is only one partial block, any point
   -- if there is no partial block, any point except the hub
   -- otherwise, any point of the partial block except the hub.

   If a partial design has a flower, removing an active point gives
   a partial design with a flower. The orbit of canonical points is
   defined to be a subset of the active points.
   A Steiner system is covered by a flower, so all ancestors also
   have flowers. Thus all the nodes in the search are partial
   designs with flowers.
*/

static boolean
Spathinvar(design *des, int s, int *inv)
/* If point s-1 is not active, return FALSE. Otherwise,
   for all i set inv[i]=0 if i is active and inv[i]=1
   if i is not active, and return TRUE. */
{
    int x,y,mb,p,q,nfull,npart;
    int i,nf,len;
    setword active,all,bitp,w;
    setword full[MAXBLOCKS],part[MAXBLOCKS];

    mb = des[0];
    x = (s-1) / (k-1);  /* number of full blocks */
    y = (s-1) % (k-1);  /* length of partial block - 1 */

    nfull = npart = 0;
    for (q = 1; q <= mb; ++q)
    {
        w = BLOCK(des[q]);
        len = POPCOUNT(w);
        if (len == k) full[nfull++] = w;
        else if (len == y+1) part[npart++] = w;
    }
    if (nfull < x) return FALSE;

    all = ALLMASK(s);

    if (s <= k)
    {
        if (nfull+npart == 0) return FALSE;
        active = all;
    }
    else
    {
        active = 0;
        for (p = 0; p < s; ++p)
        {
            bitp = bit[p];
            nf = 0;
            for (q = 0; q < nfull; ++q)
                if ((full[q]&bitp)) ++nf;
            if (nf >= x)
            {
                if (y == 0)
                    active |= (all^bitp);
                else
                {
                    for (i = 0; i < npart; ++i)
                        if ((part[i]&bitp)) active |= (part[i]^bitp);
                }
            }
        }
    }

    if (!(active&bit[s-1])) return FALSE;

    for (p = 0; p < s; ++p)
        if ((active&bit[p])) inv[p] = 0;
        else               inv[p] = 1;

    return TRUE;
}

/*******************************************************************/

static boolean
accept(design *des, int s, boolean wastriv)
/* Test if final point is canonical, and if so set autom[] and
   numautom.  wastriv=TRUE implies that des minus its last point
   had trivial group. */
{
    int lab[MAXPOINTS],ptn[MAXPOINTS],count[MAXPOINTS],orbits[MAXPOINTS];
    long invalue[MAXPOINTS];
    int pathinv[MAXPOINTS],minpathinv;
    design candes[MAXBLOCKS];
    int code,numcells,i,j,p,numvals;
    setword active;
    statsblk stats;
    static DEFAULTOPTIONS_MDES(options);
    long *inv,invi,bestinvar;
    int minnum,imin,iw,ib,ibb,tmp;
    long pointinv1,pointinv2;

    if (Steiner)
    {
        if (!Spathinvar(des,s,pathinv)) return FALSE;
    }
    else
    {
        if (!pathinvar(des,s,pathinv)) return FALSE;
    }

    numautom = 0;
    groupsize = 1;
    transitive = FALSE;
    INVAR(s);
    inv = stage[s].invar;

    /* Best invariant value is the smallest value that occurs
       the least number of times. */

    numvals = 0;
    minpathinv = pathinv[s-1]; 

    for (i = 0; i < s; ++i)
    {
        if (pathinv[i] != minpathinv) continue;
        invi = inv[i];
        for (j = 0; j < numvals; ++j)
            if (invalue[j] >= invi) break;

        if (j == numvals)
        {
            invalue[j] = invi;
            count[j] = 1;
            ++numvals;
        }
        else if (invalue[j] > invi)
        {
            for (p = numvals; --p >= j; )
            {
                invalue[p+1] = invalue[p];
                count[p+1] = count[p];
            }
            invalue[j] = invi;
            count[j] = 1;
            ++numvals;
        }
        else
            ++count[j];
    }

    minnum = MAXPOINTS+1;
  
    for (i = 0; i < numvals; ++i)
    {
        if (count[i] < minnum)
        {
            minnum = count[i];
            imin = i;
        }
    }
  
    bestinvar = invalue[imin];
    if (inv[s-1] != bestinvar) return FALSE;
    if (minnum == 1 && wastriv) return TRUE;

    iw = 0;
    ib = s - 1;
    for (i = s; --i >= 0; )
        if (pathinv[i] == minpathinv && inv[i] == bestinvar)
            lab[ib--] = i;
        else
            lab[iw++] = i;

    active = bit[0] | bit[ib+1];

    for (i = 0; i < s; ++i) ptn[i] = 1;
    ptn[s-1] = 0;
    if (ib >= 0) ptn[ib] = 0;

   /* Final cell is ib+1..s-1, size s-ib-1,
      Other cells 0..ib, size ib+1.
      lab[s-1] = s-1. */

    if (s - ib > 1 && ib >= 0)  /* non-trivial last cell */
    {
        pointinv1 = pointinvar(s,s-1,inv);
        ibb = s-2;
        for (i = s-1; --i > ib; )
        {
            pointinv2 = pointinvar(s,lab[i],inv);
            if (pointinv2 > pointinv1) return FALSE;
            if (pointinv2 == pointinv1)
            {
                tmp = lab[i];
                lab[i] = lab[ibb];
                lab[ibb] = tmp;
                --ibb;
            }
        }

       /* Now the last part is lab[ibb+1..s-1], size s-ibb-1 */
        ptn[ibb] = 0;
        active |= bit[ibb+1];
        if (wastriv && s-ibb == 2) return TRUE;   /* CHECK */

        sortindirect(lab,inv,ib+1);
        for (i = 0; i < ib; ++i)
            if (inv[lab[i+1]] != inv[lab[i]])
            {
                ptn[i] = 0;
                active |= bit[i+1];
            }
    }

    numcells = POPCOUNT(active);
  
    ++refcalls;
    refine1_des((graph*)des,lab,ptn,0,&numcells,count,&active,&code,1,s);
    if (numcells == s) return (lab[s-1] == s-1);

    if (ptn[s-2] == 0)
    {
        if (lab[s-1] != s-1) return FALSE;
        if (wastriv) return TRUE;
    }
    for (i = s-1; ; --i)
        if (lab[i] == s-1 || ptn[i-1] == 0) break;
    if (lab[i] != s-1) return FALSE;

    options.getcanon = TRUE;
    options.userautomproc = automproc;
    if (s == v) options.userlevelproc = makegroupsize;
    else        options.userlevelproc = NULL;
    options.defaultptn = FALSE;
   /* Trying active=0. */
    active = 0;
    naumdesign(des,lab,ptn,&active,orbits,&options,&stats,s,candes);
    ++nautycalls;
    if (stats.numorbits > numcells) ++reffails;
    if (stats.numorbits == 1) transitive = TRUE;

    return (orbits[s-1] == orbits[lab[s-1]]);
}

/*******************************************************************/

static boolean
feasible(int s)
/* Test if extension set is feasible for stage s */
{
    int mine,maxe;
    int q,mb,need,ex;
    multint *e,*ext,*elim;
    long numext;
    design *des;

    des = stage[s].des;
    numext = stage[s].numext;
    mb = des[0];
    ext = stage[s].ext;
    elim = ext + mb*numext;

    if (numext == 0) return FALSE;

    for (q = 1; q <= mb; ++q, ++ext)
    {
        e = ext;
	mine = maxe = *e;
        for (e += mb; e < elim; e += mb)
        {
            ex = *e;
	    if      (ex < mine) mine = ex;
	    else if (ex > maxe) maxe = ex;
        }
	need = (k - BLOCKLEN(des[q])) * MULT(des[q]);
	if ((v-s)*mine > need || (v-s)*maxe < need)
	    return FALSE;
    }

    return TRUE;
}

/******************************************************************/

static boolean
badinter(design *des, int s)
/* return TRUE if two just-extended blocks intersect in
   more than maxinter points. */
{
    setword bits,block;
    design blocks[MAXBLOCKS];
    int i,j,mb,nb,pop;

    mb = des[0];
    nb = 0;
    bits = bit[s-1];
    for (i = 1; i <= mb; ++i)
    {
        if ((des[i] & bits))
        {
            block = BLOCK(des[i]);
            pop = POPCOUNT(block);
            if (pop > maxinter && MULT(des[i]) >= 2) return TRUE;
            if (pop > maxinter+1) blocks[nb++] = block;
        }
    }

    for (j = 1; j < nb; ++j)
    {
        block = blocks[j];
        for (i = 0; i < j; ++i)
            if (POPCOUNT(block&blocks[i]) > maxinter) return TRUE;
    }

    return FALSE;
}

/******************************************************************/

static int
fullblocks(design *des, int s)
/* Number of full blocks */
{
    int mb,q,tot,max;

    max = (s > k ? k : s);
    mb = des[0];
    tot = 0;
    for (q = 1; q <= mb; ++q)
        if (BLOCKLEN(des[q]) == max) tot += MULT(des[q]);

    return tot;
}

static boolean
testolap(design *des)
/* Test if there are two blocks intersecting in olap points
   and covering all points. Only called if olap>=0 and s=2k-olap. */
{
    int mb,q,q1,q2,nfb;
    setword all,full[MAXBLOCKS];

    mb = des[0];
    all = ALLMASK(2*k-olap);

    nfb = 0;
    for (q = 1; q <= mb; ++q)
        if (BLOCKLEN(des[q]) == k)
            full[nfb++] = BLOCK(des[q]);

    for (q2 = 1; q2 < nfb; ++q2)
    for (q1 = 0; q1 < q2; ++q1)
        if ((full[q1]|full[q2]) == all) return TRUE;

    return FALSE;
}

/******************************************************************/

static void
node(int s)
/* Extend stage s to stage s+1. At entry, stage[s] is complete
   except for stage[s].eorb, and the group is in autom[].  This
   procedure will overwrite the group after using it and create
   the same state for stage s+1.  */
{
    boolean triv;
    long *eorb;
    size_t ia,ie,numext0,numext1,maxext1;
    multint *a,*e,*ext0,*e1,*e1save;
    int i,j,totlo,tothi,tot,left,mb0,mb1;
    design *des0,*des1;
    setword bits,block;
    int len[MAXBLOCKS];
    int mult0[MAXBLOCKS],lo[MAXBLOCKS],hi[MAXBLOCKS],x[MAXBLOCKS];
    int ai,ei,li,loo,hii,kml,vms;

    ++nodecount[s];

    ++totnodes;

#ifdef DEBUG
    putdesign(stdout,s,TRUE);              /* DEBUG */
    check_stage(0,s,TRUE,FALSE);           /* DEBUG */
#endif

    if (progress > 0 && totnodes % progress == 0)
    {
        for (i = 2; i < v; ++i) fprintf(stderr," %lld",nodecount[i]);
        fprintf(stderr," out=%lld(%lld)\n",nout,nsimple);
    }

    triv = (numautom == 0);
    if (!triv) { ++nontriv[s]; makeeorbits(s); }

    bits = bit[s];
    des0 = stage[s].des;
    des1 = stage[s+1].des;
    mb0 = des0[0];
    numext0 = stage[s].numext;
    if (numext0 > maxnumext) maxnumext = numext0;
    eorb = stage[s].eorb;
    ext0 = stage[s].ext;

    for (i = 0; i < mb0; ++i)
    {
        mult0[i] = MULT(des0[1+i]);
        len[i] = BLOCKLEN(des0[1+i]);
    }

    vms = v - s;
    for (ia = 0, a = ext0; ia < numext0; ++ia, a += mb0)
    {
        if (!triv && eorb[ia] != ia) continue;

#ifdef PROB
        ++avail[s];
        if (prob[s] > 1 && KRAN(prob[s]) != 0) continue;
        ++used[s];
#endif

        mb1 = 0;
        for (i = 0; i < mb0; ++i)
        {
            block = BLOCK(des0[1+i]);
            if (a[i] < mult0[i])
                des1[1+(mb1++)] = block | (mult0[i]-a[i]);
            if (a[i] > 0)
                des1[1+(mb1++)] = block | bits | a[i];
        }
        des1[0] = mb1;

        if (minfull[s+1] > 0 && fullblocks(des1,s+1) < minfull[s+1]) continue;
        if (olap >= 0 && s+1 == 2*k-olap && !testolap(des1)) continue;

        if (maxinter <= k-2 && badinter(des1,s+1)) continue;

        if (!accept(des1,s+1,triv)) continue;

        if (s == splitlevel)
        {
            if (splittest)
            {
                ++casecounter;
                continue;
            }
            if (splitcount-- != 0) continue;
            splitcount = splitmod - 1;
        }

        if (s == v-1)
        {
            got_one();
            continue;
        }

#define LB(x) if ((x)>loo) loo=(x)
#define UB(x) if ((x)<hii) hii=(x)

        e1 = stage[s+1].ext;
        numext1 = 0;
        maxext1 = stage[s+1].extsize / mb1;
        for (ie = 0, e = ext0; ie < numext0; ++ie, e += mb0)
        {
            /* Now we find all extensions of stage s compatible with a. */

            totlo = tothi = 0;

            for (i = 0; i < mb0; ++i)
            {
                ai = a[i];
                ei = e[i];
                li = len[i];

                loo = MAX(0,ai+ei-mult0[i]);
                hii = MIN(ai,ei);
                if (loo > hii) break;

                if (li == k)
                {
                    UB(0);
                }
                else if (li == k-1)
                {
                    UB(0);
                    LB(ei-maxmult);
                }
                else if (li == k-2)
                {
                    UB(maxmult);
                }
                
                kml = k - li;
                if (kml == vms)
                {
                    UB(maxmult);
                    LB(ai); UB(ai);
                    LB(ei); UB(ei);
                }
                else if (kml == vms-1)
                {
                    LB(ai-maxmult);
                    LB(ei-maxmult);
                    LB(ai+ei-mult0[i]);
                    UB(ai+ei-mult0[i]);
                } 
                else if (kml == vms-2)
                {
                    UB(maxmult+ai+ei-mult0[i]);
                }

                if (loo > hii) break;

                totlo += loo;
                tothi += hii;
                lo[i] = loo;
                hi[i] = hii - loo;
            }
  
            if (i < mb0) continue;
            if (totlo > lambda || tothi < lambda) continue;

            for (i = 0; i < mb0; ++i) x[i] = 0;

            tot = lambda - totlo;
            left = tot;
            for (i = mb0; --i >= 0; )
            {
                if (left <= hi[i])
                {
                    x[i] = left;
                    break;
                }
                else
                {
                    x[i] = hi[i];
                    left -= hi[i];
                }
            }

            for (;;)
            {
                if (numext1 >= maxext1 - 1)
                {
                    moreext(s+1);
                    e1 = stage[s+1].ext + mb1*numext1;
                    maxext1 = stage[s+1].extsize / mb1;
                }

                e1save = e1;
                for (i = 0; i < mb0; ++i)
                {
                    if (a[i] < mult0[i]) *(e1++) = e[i] - lo[i] - x[i];
                    if (a[i] > 0)        *(e1++) = lo[i] + x[i];
                }

                ++numext1;

                /* This should have made mb1 values. */

                left = x[mb0-1];
                for (i = mb0-1; --i >= 0; )
                {
                    if (x[i] < hi[i] && left > 0)
                        break;
                    else
                        left += x[i];
                }

                if (i < 0) break;

                ++x[i];
                --left;
                for (j = mb0; --j > i; )
                {
                    if (left <= hi[j])
                    {
                        x[j] = left;
                        left = 0;
                    }
                    else
                    {
                        x[j] = hi[j];
                        left -= hi[j];
                    }
                }
            }
        }
        stage[s+1].numext = numext1;

        /* Improve this to ignore forced 0 and 1. */

        if (numext1 > 0 &&
              ((b-mb1 > 6) || numext1*(1L << (b-mb1)) >= v-s-1) &&
              (s > v-4 || feasible(s+1)))
            node(s+1);
    }
}

/*******************************************************************/

static void
initstage1(void)
/* Initialize stage[1] and autom[] for single point. */
{
    numautom = 0;

    stage[1].des[0] = 2;
    stage[1].des[1] = b - r;
    stage[1].des[2] = bit[0] | r;

    stage[1].numext = 1;
    stage[1].ext[0] = r - lambda;
    stage[1].ext[1] = lambda;

    stage[1].invar[0] = 0;
}

/*******************************************************************/

static long
gcd(long n1, long n2)
/* gcd of two integers */
{
        long a,b,c;

        if (n1 < 0) n1 = -n1;
        if (n2 < 0) n2 = -n2;

        if (n1 > n2) { a = n1; b = n2; }
        else         { a = n2; b = n1; }

        while ((c = a % b) > 0) { a = b; b = c; }

        return b;
}

static boolean
quadres(long x, long y)
/* Test if x is a quadratic residue mod |y| */
{
    long u,xres,absy;

    absy = (y > 0 ? y : -y);
    xres = ((x % absy) + absy) % absy;

    for (u = 0; u < absy; ++u)
        if ((u*u) % absy == xres) return TRUE;

    return FALSE;
}

static boolean
testquad(long a, long b, long c)
/* Test if a*x^2 + b*y^2 + c*z^2 has a non-trivial integer solution.
   It can be assumed that a,b,c are not zero. */
{
    long q,w;
    boolean changed;

   /* Not all the same sign */

    q = (a < 0) + (b < 0) + (c < 0);
    if (q != 1 && q != 2) return FALSE;

   /* Make pairwise coprime */

    do
    {
        changed = FALSE;

        q = gcd(a,b);
        if (q > 1) { a /= q; b /= q; c *= q; changed = TRUE; }
        q = gcd(a,c);
        if (q > 1) { a /= q; c /= q; b *= q; changed = TRUE; }
        q = gcd(b,c);
        if (q > 1) { b /= q; c /= q; a *= q; changed = TRUE; }
    }
    while (changed);

   /* Make square-free */
 
    w = (a > 0 ? a : -a);
    for (q = 2; q*q <= w; )
        if (w % (q*q) == 0) w /= q*q; else ++q;
    a = (a > 0 ? w : -w);
 
    w = (b > 0 ? b : -b);
    for (q = 2; q*q <= w; )
        if (w % (q*q) == 0) w /= q*q; else ++q;
    b = (b > 0 ? w : -w);
 
    w = (c > 0 ? c : -c);
    for (q = 2; q*q <= w; )
        if (w % (q*q) == 0) w /= q*q; else ++q;
    c = (c > 0 ? w : -w);

   /* Quadratic residues */

    if (!quadres(-b*c,a)) return FALSE;
    if (!quadres(-a*c,b)) return FALSE;
    if (!quadres(-a*b,c)) return FALSE;

    return TRUE;
}

static boolean
issquare(long x)
/* Test if positive integer is a square */
{
    long y;

    for (y = 0; y*y < x; ++y) {}

    return y*y == x;
}

static boolean
BRC(long vv, long kk, long ll)
/* Bruck-Ryser-Chowla test; b=v is assumed */
{
    if (vv % 2 == 0) return issquare(kk-ll);

    if (((vv-1)/2) % 2 == 0) return testquad(1,ll-kk,-ll);
    else                     return testquad(1,ll-kk,ll);
}

/*******************************************************************/

static void
intscan(int *x, int v, int rhs0, int rhs1, int rhs2)
/* Variables x[v+1...] are known.
   rhs0, rhs1, rhs2 are remaining right sides of equations.
   x[mininter] is the last variable. */
{
    int i,maxi;

    if (v == mininter)
    {
        if (rhs0*v != rhs1 || rhs0*(v*v-v)/2 != rhs2)
            return;
        x[v] = rhs0;
        for (i = v; i <= maxinter; ++i)
            if (x[i] == 0) nointmask |= bit[i];
            else           yesintmask |= bit[i];
        if (x[k] > eqmultmax) eqmultmax = x[k];
        return;
    }

    maxi = rhs0;
    if (v > 0 && rhs1/v < maxi) maxi = rhs1/v;
    if (v > 1 && 2*rhs2/(v*v-v) < maxi)
        maxi = 2*rhs2/(v*v-v); 

    for (i = 0; i <= maxi; ++i)
    {
        x[v] = i;
        intscan(x,v-1,rhs0-i,rhs1-i*v,rhs2-i*(v*v-v)/2);
    }
}

static void
intsearch()
/* Solutions for intersection with full block.
   Set nointmask = intersections that sometimes don't occur
      yesintmask = intersections that sometimes do occur
         These don't have bit k by design.
   The existing mininter, maxinter and maxmult are obeyed.
*/
{
    int x[MAXPOINTS+1];
    int i,numk;

    nointmask = yesintmask = 0;
    eqmultmax = 1;

    for (numk = 1; numk <= maxmult; ++numk)
    {
        for (i = mininter; i <= maxinter; ++i) x[i] = 0;
        x[k] = numk;
        
        intscan(x,(maxinter==k?k-1:maxinter),
             b-numk,(r-numk)*k,(lambda-numk)*(k*k-k)/2);
    }

    nointmask &= ~bit[k];
    yesintmask &= ~bit[k];
}

/*******************************************************************/

static int
parameters(boolean iswitch, boolean simple)
/* Assume v, k, lambda, simple.
   Set b, r, mininter, maxinter, maxmult, simple.
   Return 0 if feasible, positive if not. */
{
    int mi,y,i,j,mini,maxi;
    long gcdbrl;

    if ((v*(v-1)*lambda) % (k*(k-1)) != 0) return 2;
    if (((v-1)*lambda) % (k-1) != 0) return 2;

    b = (v*(v-1)*lambda) / (k*(k-1));
    r = ((v-1)*lambda) / (k-1);
    if (b > MAXBLOCKS || r > MAXMULT || b-r > MAXMULT) return 7; 
         /* Too many blocks */

    if (k >= v || lambda >= r) return 1;  /* full design */
    if (b < v) return 3;  /* Fisher */

    if (b == v && !BRC(v,k,lambda)) return 5;
    if (r == k+lambda && (lambda==1 || lambda==2)
                      && !BRC(v+k+lambda,k+lambda,lambda))
        return 6;

    if (!iswitch || maxinter > k) mi = k;
    else if (iswitch && maxinter < k) mi = maxinter;
    if (simple && mi == k) mi = k-1;
    if (mi == 0) mi = 1;
    if (lambda == 1) mi = 1;

    if ((2*lambda*k + r*(r-lambda-k)) / r < mi)
        mi = (2*lambda*k + r*(r-lambda-k)) / r;   /* Connor */

    if (b == v && mi < lambda) return 4;
    if (b == v) maxinter = lambda;
    else        maxinter = mi;

    if (b == v)
        mininter = lambda;
    else if (k + lambda - r >= 0)
        mininter = k + lambda - r;   /* Connor */
    else
        mininter = 0;

    if (mininter > maxinter)
        gt_abort(">E gen2des: impossible mininter > maxinter\n");

    if (maxinter < k)  /* var Lint and Ryser */
        maxmult = 1;
    else if (lambda*v % b == 0)
        maxmult = b / v;
    else
        maxmult = (b-1) / v;

    if (maxmult*v == b)
    {
        gcdbrl = gcd(gcd(b,r),lambda);
        if (gcdbrl % maxmult != 0) --maxmult;
    }

    for (y = 1; y <= k-2; ++y)  /* Dobcsanyi, check equality case */
    {
        j = ((y+1)*y*b - 2*y*k*r + k*(k-1)*lambda)/((k-y)*(k-y-1));
        if (j < maxmult) maxmult = j;
    }

    if (maxmult == 1)
    {
        if (maxinter == k) maxinter = k-1;
        simple = TRUE;
    }

    intsearch();

    maxi = 0;
    mini = k;
    for (i = mininter; i <= maxinter; ++i)
        if ((yesintmask & bit[i]))
        {
            if (i < mini) mini = i;
            if (i > maxi) maxi = i;
        }
    if (mini > mininter) mininter = mini;
    if (maxi < maxinter) maxinter = maxi;

    if (eqmultmax < maxmult) maxmult = eqmultmax;
    if (maxmult > 1) maxinter = k;

    if (mininter > maxinter)
        gt_abort(">E gen2des: impossible mininter > maxinter\n");

    return 0;
}

/*******************************************************************/

static boolean
SRparams(int *srk, int *srlambda, int *srmu)
/* Check if blocl intersection graph is strongly regular and
   if so provide parameters. */
{
    int x,y,theta1,theta2;

    if (maxmult > 1 || (yesintmask & nointmask) 
                    || POPCOUNT(yesintmask) != 2) return FALSE;

    x = FIRSTBITNZ(yesintmask);
    y = FIRSTBITNZ(yesintmask ^ bit[x]);

    *srk = (k*(r-1)-x*(b-1)) / (y-x);
    theta1 = (r - lambda - k + x) / (y - x);
    theta2 = (x - k) / (y - x);

    *srlambda = *srk + theta1*theta2 + theta1 + theta2;
    *srmu = *srk + theta1*theta2;

    return TRUE;
}

/*******************************************************************/

int
main(int argc, char *argv[])
{
    boolean badargs;
    boolean quiet,split,verbose,mswitch,Qswitch,iswitch,Pswitch,oswitch;
    boolean pswitch,aswitch;
    int argnum,i,j,splitres;
    int ok,tmpres,tmpmod,tmplev;
    boolean hasSR;
    int srk,srlambda,srmu;
    char *arg,sw;
    double t0,t1;
    char msg[101];
    long avalue,prog,mvalue,mlevel;
    long long totnontriv;
#ifdef PROB
    long pvalue,plevel;
    double totalprob,totalest,simpleest;
#endif

    HELP; PUTVERSION;

    naumdesign_check(WORDSIZE,1,WORDSIZE-MULTBITS,NAUTYVERSIONID);

    simple = split = quiet = mswitch = uswitch = verbose = Qswitch = FALSE;
    gzip = splittest = pswitch = oswitch = iswitch = flush = Pswitch = FALSE;
    Steiner = aswitch = FALSE;
#ifdef PROB
    for (i = 0; i <= MAXPOINTS; ++i) prob[i] = 1;
#endif

    argnum = 0;
    badargs = FALSE;
    for (j = 1; !badargs && j < argc; ++j)
    {
        arg = argv[j];
        if (arg[0] == '-' && arg[1] != '\0')
        {
            ++arg;
            while (*arg != '\0')
            {
                sw = *arg++;
                     SWBOOLEAN('s',simple)
                else SWINT('i',iswitch,maxinter,"gen2des -i")
                else SWINT('o',oswitch,olap,"gen2des -o")
                else SWBOOLEAN('S',Steiner)
                else SWLONG('a',aswitch,avalue,"gen2des -a")
                else SWBOOLEAN('q',quiet)
                else SWBOOLEAN('u',uswitch)
                else SWBOOLEAN('v',verbose)
                else SWBOOLEAN('Q',Qswitch)
                else SWBOOLEAN('L',splittest)
                else SWBOOLEAN('z',gzip)
                else SWLONG('P',Pswitch,prog,"gen2des -P")
                else SWBOOLEAN('F',flush)
                else SWRANGE('m',"/",mswitch,mvalue,mlevel,"gen2des -m")
#ifdef PROB
                else SWRANGE('p',"/",pswitch,pvalue,plevel,"gen2des -p")
#endif
                else badargs = TRUE;

#ifdef PROB
                if (pswitch)
                {
                    if (plevel < 3 || plevel > MAXPOINTS || pvalue < 1)
                        badargs = TRUE;
                    else
                        prob[plevel] = pvalue;
                    pswitch = FALSE;
                }
#endif
                if (mswitch)
                {
                    if (mlevel < 2 || mlevel > MAXPOINTS
                                   || mvalue < 0 || mvalue > MAXBLOCKS)
                        badargs = TRUE;
                    else
                        minfull[mlevel] = mvalue;
                    mswitch = FALSE;
                }
            }
        }
        else if (sscanf(arg,"%d/%d/%d",&tmpres,&tmpmod,&tmplev) == 3)
        {
            split = TRUE;
            splitres = tmpres;
            splitmod = tmpmod;
            splitlevel = tmplev;
        }
        else
        {
            ++argnum;
            if (argnum == 1)
            {
                if (sscanf(arg,"%d",&v) != 1 || v <= 2)
                    gt_abort(">E gen2des: bad value of v\n");
                if (v > MAXPOINTS)
                    gt_abort(">E gen2des: too many points\n");
            }
            else if (argnum == 2)
            {
                if (sscanf(arg,"%d",&k) != 1 || k <= 1 || k >= v-1)
                    gt_abort(">E gen2des: bad value of k\n");
            }
            else if (argnum == 3)
            {
                if (sscanf(arg,"%d",&lambda) != 1 || lambda < 1)
                    gt_abort(">E gen2des: bad value of lambda\n");
            }
            else                  badargs = TRUE;
        }
    }

    if (badargs || argnum != 3)
    {
        fprintf(stderr,">E Usage: %s\n",USAGE);
        GETHELP;
        exit(1);
    }

#ifndef ZLIB
    if (gzip) gt_abort(
         ">E gen2des: -z is only available if compiled with -DZLIB\n");
#endif
    if (gzip && flush) gt_abort(">E gen2des: -z and -f are incompatible\n");
    if (gzip && isatty(fileno(stdout))) gt_abort(
       ">E gen2des: probably you don't want gzip output to your terminal\n");

    if (!oswitch) olap = -1;

    if (oswitch && Steiner) gt_abort(">E -m, -o -S are incompatible\n");
    for (i = 0; i <= v; ++i)
        if (minfull[i] > 0 && (oswitch || Steiner))
            gt_abort(">E -m, -o -S are incompatible\n");
    if (Steiner && lambda != 1)
        gt_abort(">E -S is only allowed for lambda=1\n");

    for (i = 0; i <= k; ++i) if (minfull[i] == 0) minfull[i] = 1;

    if (!split)
    {
        splitres = 0;
        splitmod = 1;
        splitlevel = 0;
    }

    if (!Pswitch || prog < 0)
        progress = 0;
    else if (prog == 0)
        progress = 1;
    else
        progress = prog * 1000LL;

    if (splitmod <= 0 || splitres < 0 || splitres >= splitmod)
        gt_abort(">E gen2des: bad res/mod values\n");

    if (v <= 5 && splitmod != 1)
        gt_abort(">E gen2des: splitting not available for v <= 5\n");

    ok = parameters(iswitch,simple);
    if (ok > 0)
    {
        if      (ok == 1)
            gt_abort(">E gen2des: k >= v or lambda >= r\n");
        else if (ok == 2)
            gt_abort(">E gen2des: infeasible parameters (divisibility)\n");
        else if (ok == 3)
            gt_abort(">E gen2des: infeasible parameters (Fisher)\n");
        else if (ok == 4)
            gt_abort(">E gen2des: infeasible parameters (-i too small)\n");
        else if (ok == 5)
            gt_abort(
                ">E gen2des: infeasible parameters (Bruck-Ryser-Chowla)\n");
        else if (ok == 6)
            gt_abort(">E gen2des: infeasible parameters (residual BRC)\n");
        else if (ok == 7)
            gt_abort(
               ">E gen2des: too many blocks, r > MAXMULT or b-r > MAXMULT\n");
    }

    if (b > MAXBLOCKS)
        gt_abort(">E gen2des: too many blocks\n");

    if (oswitch && olap < 0 || olap > v-1)
        gt_abort(">E gen2des: -o argument must be 0..v-1\n");
    if (oswitch && ((nointmask&bit[olap]) || !(yesintmask&bit[olap])))
        fprintf(stderr,
                ">W  Warning: -o argument is not a compulsory overlap\n");

    if (!aswitch || avalue < 1) minautoms = 1; else minautoms = avalue;

    hasSR = SRparams(&srk,&srlambda,&srmu);

    if (!quiet || Qswitch)
    {
        msg[0] = '\0';
        CATMSG0(">A gen2des");
        if (simple) CATMSG0(" -s");
        if (Steiner) CATMSG0(" -S");
        if (uswitch) CATMSG0(" -u");
        if (minautoms>1) CATMSG1(" -a%llu",minautoms);
        if (olap>=0) CATMSG1(" -o%d",olap);
        if (split) CATMSG3(" %d/%d/%d",splitres,splitmod,splitlevel);        
        CATMSG5(" v=%d k=%d lambda=%d r=%d b=%d",
                v,k,lambda,r,b);
        CATMSG3(" mult<=%d int=%d-%d",maxmult,mininter,maxinter);
        for (i = mininter; i <= maxinter; ++i)
            if ((yesintmask & bit[i]))
            {
                CATMSG0(i == mininter ? "(" : ",");
                if (!(nointmask & bit[i])) CATMSG1("+%d",i);
                else                       CATMSG1("%d",i);
            }
        CATMSG0(")");
        if (hasSR)
        {
            CATMSG4(" SR(%d,%d,%d,%d)",b,srk,srlambda,srmu);
            CATMSG4(",SR(%d,%d,%d,%d)",
                    b,b-srk-1,b-2-2*srk+srmu,b-2*srk+srlambda);
        }

        fprintf(stderr,"%s\n",msg);
#ifdef PROB
        totalprob = 1.0;
        for (i = 0; i <= v-1; ++i) totalprob *= prob[i];
        CATMSG1(" prob=%.8e",totalprob);
#endif
    }

    if (Qswitch) exit(0);

#ifdef PROB
    INITRANBYTIME;
#endif

    numgroups = 0;
    for (i = 1; i < v; ++i)
    {
        if ((stage[i].ext = malloc(8192*sizeof(multint))) == NULL)
            gt_abort(">E gen2des: initial malloc failed\n");
        stage[i].extsize = 8192;
        stage[i].eorb = NULL;
        stage[i].eorbsize = 0;
    }

    sprintf(line,"%d %d",v,b);
    for (linepos = line; *linepos != '\0'; ++linepos) {}
    splitcount = splitres;

#ifdef ZLIB
    if (gzip)
    {
        gzfile = gzdopen(fileno(stdout),"w");
        if (gzfile == NULL)
            gt_abort(">E gen2des: failed to open gzip stream to stdout\n");
    }
#endif 

    t0 = CPUTIME;

    initstage1();
    node(1);

    t1 = CPUTIME;

#ifdef ZLIB
    if (gzip)
    {
        if (gzclose(gzfile) != Z_OK)
            gt_abort(">E gen2des: Error on closing gzip stream\n");
    }
#endif

    if (splittest)
    {
        fprintf(stderr,">Z %lld cases at level %d in %.2f sec\n",
                   casecounter,splitlevel,t1-t0);
        exit(0);
    }

    totnontriv = 0;
    for (i = 1; i < v; ++i) totnontriv += nontriv[i];

    if (verbose)
    {
        fprintf(stderr,
                ">G   |Aut(D)|    designs     simple  trans simtrans\n");
        for (i = 0; i < numgroups; ++i)
            fprintf(stderr,">G %10llu %10lld %10lld %5lld %5lld\n",
                groupcount[i].groupsize,
                groupcount[i].totalnum,groupcount[i].simplenum,
                groupcount[i].trans,groupcount[i].simpletrans);

        fprintf(stderr,">C");
        for (i = 1; i < v; ++i)
            fprintf(stderr," %lld",nodecount[i]);
        fprintf(stderr,"\n");
    }
    fprintf(stderr,">S %lld nodes, %lld nontriv; "
                "%lld refine, %lld nauty, %lld reffails; maxext %lld\n",
                totnodes,totnontriv,refcalls,nautycalls,reffails,maxnumext);
#if PROB
    totalest = nout;
    simpleest = nsimple;
    for (i = 2; i < v; ++i)
    {
        if (nout > 0) totalest *= (double)avail[i] / (double)used[i];
        if (nsimple > 0) simpleest *= (double)avail[i] / (double)used[i];
    }

    fprintf(stderr,">Q Estimate: %.2e (%.2e simple)\n",totalest,simpleest);
#endif

    if (minautoms > 1 && !uswitch)
        fprintf(stderr,
           ">N %lld designs with non-trivial groups written\n",numwritten);
    fprintf(stderr,">Z %lld designs generated (%lld simple); %.2f sec\n",
                   nout,nsimple,t1-t0);

    exit(0);
}
