/*****************************************************************************
*                                                                            *
*  Design-specific source file for version 2.8 of nauty.                     *
*                                                                            *
*   Copyright (1984-2023) Brendan McKay.  All rights reserved.               *
*   Subject to waivers and disclaimers in COPYRIGHT.                         *
*                                                                            *
*****************************************************************************/

#include "naumdesign.h"

    /* macros for hash-codes: */
#define MASH(l,i) ((((l) ^ 065435) + (i)) & 077777)
    /* : expression whose long value depends only on long l and int/long i.
         Anything goes, preferably non-commutative. */

#define CLEANUP(l) ((unsigned int)((l) % NAUTY_INFINITY))
    /* : expression whose value depends on long l and is less than
         NAUTY_INFINITY when converted to int then short.  Anything goes. */

dispatchvec dispatch_mdesign =
  {isautom_des,testcanlab_des,updatecan_des,refine_des,refine1_des,
   cheapautom_des,targetcell_des,naumdesign_freedyn,naumdesign_check,NULL,NULL};

#define ACCUM(x,y)   x = (((x) + (y)) & 077777)

static int fuzz1[] = {037541,061532,005257,026416};
static int fuzz2[] = {006532,070236,035523,062437};

#define FUZZ1(x) ((x) ^ fuzz1[(x)&3])
#define FUZZ2(x) ((x) ^ fuzz2[(x)&3])

/*****************************************************************************
*                                                                            *
*  shell_des(x,b) sorts the multidesign x[0..b-1]                            *
*         (without initial count) using shellsort                            *
*                                                                            *
*****************************************************************************/

void
shell_des(design *x, int b)
{
    int i,j,h;
    design iw1;

    j = b / 3;
    h = 1;
    do
        h = 3 * h + 1;
    while (h < j);

    do
    {
        for (i = h; i < b; ++i)
        {
            iw1 = x[i];
            for (j = i; x[j-h] > iw1; )
            {
                x[j] = x[j-h];
                if ((j -= h) < h) break;
            }
            x[j] = iw1;
        }
        h /= 3;
    }
    while (h > 0);
}

/*****************************************************************************
*                                                                            *
*  sortindirect(x,n,z) - sort x[0..n-1] according to z[x[i]] values          *
*                                                                            *
*****************************************************************************/

static void
sortindirect(int *x, int n, unsigned int *z)
{
    int i,j,h,xi;
    unsigned int iw;

    j = n / 3;
    h = 1;
    do
        h = 3 * h + 1;
    while (h < j);

    do
    {
        for (i = h; i < n; ++i)
        {
            xi = x[i];
            iw = z[xi];
            for (j = i; z[x[j-h]] > iw; )
            {
                x[j] = x[j-h];
                if ((j -= h) < h) break;
            }
            x[j] = xi;
        }
        h /= 3;
    }
    while (h > 0);
}

/*****************************************************************************
*                                                                            *
*  isautom_des(g,perm,digraph,m,n) = TRUE iff perm is an automorphism of g   *
*  (i.e., g^perm = g).  digraph and m are ignored.                           *
*                                                                            *
*****************************************************************************/

boolean
isautom_des(graph *g, int *perm, boolean digraph, int m, int n)
{
    int i,b;
    design *des,*des0,*des1,*dlo,*dhi,*dmid,newblock,w;

    des = (design*)g;
    b = des[0];
    if (b == 0) return TRUE;

    des0 = des + 1;   /* first block */
    des1 = des + b;  /* last block */

    for (des = des0; des <= des1; ++des)
    {
        w = BLOCK(*des);
        newblock = MULT(*des);

        while (w)
        {
            TAKEBIT(i,w);
            newblock |= bit[perm[i]];
        }
        
        dlo = des0;
        dhi = des1;

        while (dlo <= dhi)
        {
            dmid = dlo + (dhi-dlo)/2;
            if (*dmid == newblock)
                break;
            else if (*dmid < newblock)
                dlo = dmid + 1;
            else
                dhi = dmid - 1;
        }
        if (dlo > dhi) return FALSE;
    }

    return TRUE;
}

/*****************************************************************************
*                                                                            *
*  blockautom(des,vperm,bperm,n) finds the block permutation bperm           *
*  induced by the point automorphism vperm. Blocks are numbered from 0.      *
*                                                                            *
*****************************************************************************/

void
blockautom(design *des, int *vperm, int *bperm, int n)
{
    int i,b;
    design *des0,*des1,*dlo,*dhi,*dmid,newblock,w;

    b = des[0];

    des0 = des + 1;   /* first block */
    des1 = des + b;  /* last block */

    for (des = des0; des <= des1; ++des)
    {
        w = BLOCK(*des);
        newblock = MULT(*des);

        while (w)
        {
            TAKEBIT(i,w);
            newblock |= bit[vperm[i]];
        }
        
        dlo = des0;
        dhi = des1;

        while (dlo <= dhi)
        {
            dmid = dlo + (dhi-dlo)/2;
            if (*dmid == newblock)
                break;
            else if (*dmid < newblock)
                dlo = dmid + 1;
            else
                dhi = dmid - 1;
        }
        bperm[des-des0] = (int)(dmid - des0);
    }
}

/*****************************************************************************
*                                                                            *
*  testcanlab_des(g,canong,lab,samerows,m,n) compares g^lab to canong,       *
*  using an ordering which is immaterial since it's only used here.  The     *
*  value returned is -1,0,1 if g^lab <,=,> canong.  *samerows is set to      *
*  the number of rows (0..n) of canong which are the same as those of g^lab. *
*                                                                            *
*****************************************************************************/

int
testcanlab_des(graph *g, graph *canong, int *lab, int *samerows, int m, int n)
{
    int i,j;
    design *des,*cdes,newblock,w;
    int b;
    design permbit[WORDSIZE];
    design workdes[MAXBLOCKS+1]; 

    des = (design*)g;
    cdes = (design*)canong;
    b = des[0];

    *samerows = 0;

    for (i = 0; i < n; ++i) permbit[lab[i]] = bit[i];

    for (i = 1; i <= b; ++i)
    {
        newblock = MULT(des[i]);
        w = BLOCK(des[i]);
        while (w)
        {
            TAKEBIT(j,w);
            newblock |= permbit[j];
        }
        workdes[i] = newblock;
    }

    shell_des(workdes+1,b);

    for (i = 1; i <= b; ++i)
    {
        if (workdes[i] > cdes[i]) return -1;
        else if (workdes[i] < cdes[i]) return 1;
    }

    *samerows = n;
    return 0;
}

/*****************************************************************************
*                                                                            *
*  updatecan_des(g,canong,lab,samerows,m,n) sets canong = g^lab,             *
*  ignoring the value of samerows and m.                                     *
*                                                                            *
*****************************************************************************/

void
updatecan_des(graph *g, graph *canong, int *lab, int samerows, int m, int n)
{
    int i,j,b;
    design *des,*cdes,newblock,w;
    design permbit[WORDSIZE];

    des = (design*)g;
    cdes = (design*)canong;
    b = des[0];
    cdes[0] = b;

    for (i = 0; i < n; ++i) permbit[lab[i]] = bit[i];

    for (i = 1; i <= b; ++i)
    {
        newblock = MULT(des[i]);
        w = BLOCK(des[i]);
        while (w)
        {
            TAKEBIT(j,w);
            newblock |= permbit[j];
        }
        cdes[i] = newblock;
    }

    shell_des(cdes+1,b);
}

/*****************************************************************************
*                                                                            *
*  refine_des(g,lab,ptn,level,numcells,count,active,code,m,n) performs a     *
*  refinement operation on the partition at the specified level of the       *
*  partition nest (lab,ptn).  *numcells is assumed to contain the number of  *
*  cells on input, and is updated.  The initial set of active cells (alpha   *
*  in the paper) is specified in the set active.  Precisely, x is in active  *
*  iff the cell starting at index x in lab is active.                        *
*  The resulting partition is equitable if active is correct (see the paper  *
*  and the Guide).                                                           *
*  *code is set to a value which depends on the fine detail of the           *
*  algorithm, but which is independent of the labelling of the design.       *
*  count is used for work space.                                             *
*                                                                            *
*****************************************************************************/

void
refine_des(graph *g, int *lab, int *ptn, int level, int *numcells,
           int *count, set *active, int *code, int m, int n)
{
    refine1_des(g,lab,ptn,level,numcells,count,active,code,m,n);
}

/*****************************************************************************
*                                                                            *
*  refine1_des(g,lab,ptn,level,numcells,count,active,code,m,n) is a version  *
*  of refine_des that works only for m=1.  It gives the same results but     *
*  is hopefully faster.                                                      *
*                                                                            *
*****************************************************************************/

void
refine1_des(graph *g, int *lab, int *ptn, int level, int *numcells,
           int *count, set *active, int *code, int m, int n)
{
    int cell1,cell2,iw,pw;
    boolean same;
    unsigned long longcode;
    unsigned int weight[WORDSIZE];
    int b,k;
    setword act,acell;
    setword block;
    int i,j,iact,bcode;
    design *des,*d;
    int workperm[WORDSIZE],workvec[WORDSIZE];

    longcode = *numcells;
    des = (design*)g;
    b = des[0];
    ++des;
    act = *active;

    j = FUZZ1(0);
    for (i = 0; i < n; ++i)
    {
        weight[i] = 0;
        workvec[lab[i]] = j;
        if (ptn[i] <= level) j = FUZZ1(i+1);
    }

    while (*numcells < n && act)
    {
        TAKEBIT(iact,act);
        longcode = MASH(longcode,iact);

        i = iact;
        acell = bit[lab[i]];
        while (ptn[i] > level) acell |= bit[lab[++i]];

        for (d = des+b; --d >= des;)
        {
            block = BLOCK(*d);
            if ((block & acell) == 0) continue;
            k = 0;
            bcode = 0;
            while (block)
            {
                TAKEBIT(i,block);
                ACCUM(bcode,workvec[i]);
                workperm[k++] = i;
            }
            bcode = (*d & MULTMASK) * (FUZZ2(bcode) ^ k);
            for (i = 0; i < k; ++i)
                ACCUM(weight[workperm[i]],bcode);
        }

        for (cell1 = 0; cell1 < n; cell1 = cell2 + 1)
        {
            pw = weight[lab[cell1]];
            same = TRUE;
            for (cell2 = cell1; ptn[cell2] > level; ++cell2)
                if (weight[lab[cell2+1]] != pw) same = FALSE;
            if (same) continue;

            if (cell2 <= cell1 + 10)
                for (i = cell1 + 1; i <= cell2; ++i)
                {
                    iw = lab[i];
                    pw = weight[iw];
                    for (j = i; weight[lab[j-1]] > pw; )
                    {
                        lab[j] = lab[j-1];
                        if (--j <= cell1) break;
                    }
                    lab[j] = iw;
                }
            else
                sortindirect(lab+cell1,cell2-cell1+1,weight);

            j = FUZZ1(cell1);
            for (i = cell1 + 1; i <= cell2; ++i)
            {
                if (weight[lab[i]] != weight[lab[i-1]])
                {
                    ptn[i-1] = level;
                    ++*numcells;
                    act |= bit[i];
                    longcode = MASH(longcode,i+weight[lab[i]]);
                    j = FUZZ1(i);
                }
                workvec[lab[i]] = j;
            }
        }
    }

    longcode = MASH(longcode,*numcells);
    *code = CLEANUP(longcode);
}

/*****************************************************************************
*                                                                            *
*  cheapautom_des(ptn,level,digraph,n) returns TRUE if the partition at the  *
*  specified level in the partition nest (lab,ptn) {lab is not needed here}  *
*  satisfies a simple sufficient condition for its cells to be the orbits of *
*  some subgroup of the automorphism group.  Otherwise it returns FALSE.     *
*  It always returns FALSE if digraph!=FALSE.                                *
*                                                                            *
*  nauty assumes that this function will always return TRUE for any          *
*  partition finer than one for which it returns TRUE.                       *
*                                                                            *
*****************************************************************************/

boolean
cheapautom_des(int *ptn, int level, boolean digraph, int n)
{
    return FALSE;
}

/*****************************************************************************
*                                                                            *
*  naumdesign_check() checks that this file is compiled compatibly with the  *
*  given parameters.   If not, call exit(1).                                 *
*                                                                            *
*****************************************************************************/

void
naumdesign_check(int wordsize, int m, int n, int version)
{
    if (wordsize != WORDSIZE)
    {
        fprintf(ERRFILE,"Error: WORDSIZE mismatch in naumdesign.c\n");
        exit(1);
    }

    if (m > 1)
    {
        fprintf(ERRFILE,"Error: only m=1 is supported by naumdesign.c\n");
        exit(1);
    }

    if (n > WORDSIZE-MULTBITS)
    {
        fprintf(ERRFILE,"Error: n too big for naumdesign.c\n");
        exit(1);
    }

    if (version < NAUTYREQUIRED)
    {
        fprintf(ERRFILE,"Error: naumdesign.c version mismatch\n");
        exit(1);
    }

    nauty_check(wordsize,m,n,version);
}

/*****************************************************************************
*                                                                            *
*  naumdesign_freedyn() - free the dynamic memory in this module (none)      *
*                                                                            *
*****************************************************************************/

void
naumdesign_freedyn(void)
{
}

/*****************************************************************************
*                                                                            *
*  bestcell_des(g,lab,ptn,level,tc_level,m,n) returns the index in           *
*  lab of the start of the "best non-singleton cell" for fixing.             *
*  If there is no non-singleton cell it returns n.                           *
*  This implementation finds the first smallest nontrivial cell,             *
*  which is not really good but something to work with.                      *
*                                                                            *
******************************************************************************/

static int
bestcell_des(graph *g, int *lab, int *ptn, int level, int tc_level,
         int m, int n)
{
    int i,j,k;
    int smallsize,size;

    smallsize = n+1;
    i = 0;
    while (i < n)
    {
        if (ptn[i] > level)
        {
            j = i;
            while (ptn[i] > level) ++i;
            size = i-j+1;
            if (size < smallsize)
            {
                smallsize = size;
                k = j;
            }
        }
        ++i;
    }

    if (smallsize <= n) return k;
    else                return n;
}

/*****************************************************************************
*                                                                            *
*  targetcell(g,lab,ptn,level,tc_level,digraph,hint,m,n) returns the index   *
*  in lab of the next cell to split.                                         *
*  hint is a suggestion for the answer, which is obeyed if it is valid.      *
*  Otherwise we use bestcell() up to tc_level and the first non-trivial      *
*  cell after that.                                                          *
*                                                                            *
*****************************************************************************/

int
targetcell_des(graph *g, int *lab, int *ptn, int level, int tc_level,
           boolean digraph, int hint, int m, int n)
{
    int i;

    if (hint >= 0 && ptn[hint] > level &&
                         (hint == 0 || ptn[hint-1] <= level))
        return hint;
    else if (level <= tc_level)
        return bestcell_des(g,lab,ptn,level,tc_level,m,n);
    else
    {
        for (i = 0; i < n && ptn[i] <= level; ++i) {}
        return (i == n ? 0 : i);
    }
}

/*****************************************************************************
*                                                                            *
*  naumdesign calls nauty with a multidesign. n is the nnumber of points.    *
*  lab, ptn, active, orbits, options and stats have the same meaning as      *
*  for nauty(). If stats->getcanon=TRUE, h[] is set to a canonical version   *
*  of the design; otherwise h=NULL is permitted.                             *
*                                                                            *
*****************************************************************************/

void
naumdesign(design *g, int *lab, int *ptn, setword *active, int *orbits,
           optionblk *options, statsblk *stats, int n, design *h)
{
    setword dnwork[1000];

    if (options->dispatch != &dispatch_mdesign)
    {
        fprintf(ERRFILE,"Error: mdesigns needs standard options block\n");
        exit(1);
    }

    nauty((graph*)g,lab,ptn,active,orbits,options,stats,dnwork,
         1000,1,n,(graph*)h);
}
