/* Perform operations on 2-designs.
   Version 1.0, June 2023.  Brendan McKay.
   Subject to the same licence as nauty 2.8.

   To compile this program you need naumdesign.h and naumdesign.c
   as well as nauty version 2.8 or later. You can compile it as
   32-bit or 64-bit.  The maximum number of points is 24 for the
   32-bit version and 56 for the 64-bit version. The maximum
   number of blocks is 1024 for both versions, but no block
   can have multiplicity more than 255.

   To activate the code for reading and writing files in gzip
   format, compile with -DZLIB and link with -lz.

   To activate the code for writing SHA1 certificates, compile with
   -DSHA and link with -lcrypto. This code uses the SHA1 function
   that comes with OpenSSL. In Debian/Ubuntu you need the package
   libssl-dev.
*/

/* Notes for future use:
   Magic number for gzipped files is 1f 8b
*/

#include "gtools.h"
#include "naumdesign.h"
#include <ctype.h>
#ifdef ZLIB
#include <zlib.h>
#endif
#ifdef SHA
#include <openssl/sha.h>
#endif

#ifdef ZLIB
#define ZLIBHELP \
"  Input files with name ending in .gz are assumed to be in gzip\n" \
"    format. To gzip the output file, use a name ending in .gz\n" \
"    or add the switch -z.\n"
#define ZLIBUSE "|-z"
#else
#define ZLIBHELP ""
#define ZLIBUSE ""
#endif

#ifdef SHA
#define SHAHELP \
"     -h   Write a 20-byte canonical hash code\n" 
#define SHAUSE "|-h"
#else
#define SHAHELP ""
#define SHAUSE ""
#endif

#define USAGE \
  "proc2des [-u" ZLIBUSE "] [-l] [-s] [-a#:#] [-p#:#] [-f] " \
     "[-m|-M|-w|-B|-b#:# " SHAUSE "]\n      " \
     " [-c] [-d] [-R|-D] [-v|-V]  [infile [outfile]]"

#define HELPTEXT \
"  Process 2-designs\n" \
"\n" \
"     -v   Give table of counts by automorphism group\n" \
"     -V   As -v but include cases excluded by -a or -o\n" \
"     -u   No designs are written\n" \
"     -l   Canonically label outputs\n" \
SHAHELP \
"     -s   Restrict output to simple designs\n" \
"     -c   Take block-wise complements of input designs\n" \
"     -d   Take duals of symmetric designs\n" \
"     -R   Make residuals of symmetric designs instead\n" \
"     -D   Make derivatives of symmetric designs instead\n" \
"     -m   Write as a v*b matrix\n" \
"     -M   Write as a b*v matrix\n" \
"     -w   Write as a list of blocks\n" \
"     -B   Write the point-block incidence graph in sparse6 format\n" \
"     -b#:#  Write the block intersection graph for overlaps #-#\n" \
"     -p#:#  Only process these inputs (first input is number 1)\n" \
"     -f   With -p, assume all input lines have the same length\n" \
"     -a#:#  Restrict automorphism group order\n" \
"     -o#:#  Restrict number of orbits on points\n" \
"\n" \
"  infile : The input file. If it is missing or '-' stdin is used.\n" \
"  outfile : The input file. If it is missing or '-' stdin is used.\n" \
ZLIBHELP

#ifndef INVAR
#define INVAR tripinv
#endif

static boolean simple,uswitch;
static boolean hasextra;

typedef struct paramblock
{
    int v,b,r,k,lambda;
    boolean valid,simple;  /* valid means is a 2-design */
} paramblock;

static long long nout,nsimple;

#define LINELEN (1000+MAXBLOCKS*(1+(MAXPOINTS+3)/4))
static char line[LINELEN];
extern size_t s6len;

static unsigned long long groupsize; 
#define MAXGROUPS 200   /* Maximum number of different group sizes */
static struct
{
    unsigned long long groupsize;
    long long totalnum,simplenum;
    long long trans,simpletrans;
} groupcount[MAXGROUPS];
static int numgroups;

#define MAXPARAMS 500  /* Maximum number of different parameter sets */
static struct
{
    long long totalnum,simplenum;
    int v,b,r,k,lambda;
    boolean valid;
} paramcount[MAXPARAMS];
static int numparams;

#ifdef ZLIB
static gzFile infilegz,outfilegz;
#endif
static boolean gzipin,gzipout;
static FILE *infile,*outfile;

/*******************************************************************/

/* Two procedures for debugging purposes */

static void
dummy(int id, int s, char *msg)
{
    printf(" ");
}

static unsigned long long
deshash(design *des, int s)
/* Hash code for design, intended for debugging */
{
    int q,mb;
    unsigned long long ans;

    mb = des[0];
    ans = s;

    for (q = 1; q <= mb; ++q)
        ans = 73 * ans + des[q];

    return ans;
}

/*******************************************************************/

static void
writedes(design *des, paramblock *param)
{
    int i,j,q,mb,i0,m;
    static unsigned char hex[]
       = {'0','1','2','3','4','5','6','7','8','9','a','b','c','d','e','f'};
    char *pos;
    setword w;

    ++nout;
    if (uswitch) return;

    mb = des[0];

    sprintf(line,"%d %d",param->v,param->b);
    for (pos = line; *pos != '\0'; ++pos) {}

    i0 = 4 * ((param->v+3)/4 - 1);
    for (q = 1; q <= mb; ++q)
    {
        w = des[q] >> (WORDSIZE - param->v);
        m = MULT(des[q]);
        for (j = 0; j < m; ++j)
        {
            *(pos++) = ' ';
            for (i = i0; i >= 0; i -= 4) *(pos++) = hex[(w>>i)&0xf];
        }
    }
    *(pos++) = '\n';

    if (gzipout)
    {
#ifdef ZLIB
        gzwrite(outfilegz,line,pos-line);
#endif
    }
    else
        fwrite(line,1,pos-line,outfile);
}

static void
writematrix(long pos, design *des, int v, int b, boolean bypoint)
/* Write an incidence matrix to outfile/outfilegz.  If bypoint=TRUE,
   then write v*b matrix. Otherwise write a b*v matrix. */
{
    char buff[MAXPOINTS+MAXBLOCKS],*p,c;
    int mb,i,l,q,m;

    mb = des[0];

    sprintf(buff,"Design %ld v=%d b=%d\n",pos,v,b);
    if (gzipout)
    {
#ifdef ZLIB
        gzwrite(outfilegz,buff,strlen(buff));
#endif
    }
    else
        fwrite(buff,1,strlen(buff),outfile);
    
    if (bypoint)
    {
        for (i = 0; i < v; ++i)
        {
            p = buff;
            for (q = 1; q <= mb; ++q)
            {
                if ((des[q]&bit[v-1-i])) c = '1'; else c = '0';
                m = MULT(des[q]);
                for (l = 0; l < m; ++l) *(p++) = c;
            }
            *(p++) = '\n';
            *p = '\0';
            if (gzipout)
            {
#ifdef ZLIB
                gzwrite(outfilegz,buff,p-buff);
#endif
            }
            else
                fwrite(buff,1,p-buff,outfile);
        }
    }
    else
    {
        for (q = 1; q <= mb; ++q)
        {
            p = buff;
            for (i = 0; i < v; ++i)
                if ((des[q]&bit[v-1-i])) *(p++) = '1'; else *(p++) = '0';
            m = MULT(des[q]);
            *(p++) = '\n';
            *p = '\0';
            if (gzipout)
            {
#ifdef ZLIB
                for (l = 0; l < m; ++l) gzwrite(outfilegz,buff,p-buff);
#endif
            }
            else
                for (l = 0; l < m; ++l) fwrite(buff,1,p-buff,outfile);
        }
    }

    ++nout;
}

static void
writeblocks(long pos, design *des, int v, int b)
/* Write a list of blocks to infile or infilegz. */
{
    int mb,q,i,m;
    setword w,hibit;
    char buff[60];
    boolean first;

    mb = des[0];

    sprintf(buff,"Design %ld v=%d b=%d\n",pos,v,b);
    if (gzipout)
    {
#ifdef ZLIB
        gzwrite(outfilegz,buff,strlen(buff));
#endif
    }
    else
        fwrite(buff,1,strlen(buff),outfile);

    for (q = 1; q <= mb; ++q)
    {
        w = BLOCK(des[q]);
        m = MULT(des[q]);

        if (gzipout)
        {
#ifdef ZLIB
            first = TRUE;
            while (w)
            {
                REMOVEHIBIT(hibit,w);
                i = v - 1 - FIRSTBITNZ(hibit);
                if (first) gzprintf(outfilegz,"%d",i);
                else       gzprintf(outfilegz," %d",i);
                first = FALSE;
            }
            if (m > 1) gzprintf(outfilegz," * %d\n",m);
            else       gzprintf(outfilegz,"\n");
#endif
        }
        else
        {
            first = TRUE;
            while (w)
            {
                REMOVEHIBIT(hibit,w);
                i = v - 1 - FIRSTBITNZ(hibit);
                if (first) fprintf(outfile,"%d",i);
                else       fprintf(outfile," %d",i);
                first = FALSE;
            }
            if (m > 1) fprintf(outfile," * %d\n",m);
            else       fprintf(outfile,"\n");
        }
    }

    ++nout;
}

static void
writeincidence(design *des, int v, int b)
/* Write the bipartite incidence graph, points first. */
{
    int mb,q,i,j,t,m,n,ne,ie,deg;
    setword w;
    static SG_DECL(sg);
    int *ee,*dd;
    size_t *vv;
    char *s6;

    mb = des[0];
    ne = 0;
    for (q = 1; q <= mb; ++q)
        ne += MULT(des[q]) * BLOCKLEN(des[q]);

    n = v + b;
    SG_ALLOC(sg,v+b,2*ne,"proc2des");
    SG_VDE(&sg,vv,dd,ee);
    sg.nv = n;
    sg.nde = 2*ne;

    ie = 0;
    for (i = 0; i < v; ++i)
    {
        deg = 0;
        vv[i] = ie;
        w = bit[i];
        j = v;
        for (q = 1; q <= mb; ++q)
            if ((des[q] & w)) 
            {
                m =  MULT(des[q]);
                for (t = 0; t < m; ++t) ee[ie++] = j++;
                deg += m;
            }
        dd[i] = deg;
     }
      
     for (q = 1; q <= mb; ++q)
     {
        m = MULT(des[q]);
        for (t = 0; t < m; ++t)
        {
            w = BLOCK(des[q]);
            dd[i] = POPCOUNT(w);
            vv[i] = ie;
            ++i;
            while (w)
            {
                TAKEBIT(j,w);
                ee[ie++] = j;
            }
        }
    }

    ++nout;
    
    s6 = sgtos6(&sg);

    if (gzipout)
    {
#ifdef ZLIB
        gzwrite(outfilegz,s6,s6len);
#endif
    }
    else
        fwrite(s6,1,s6len,outfile);
}

static void
writebigraph(design *des, int v, int b, long lo, long hi)
/* Write graph whose vertices are the blocks and whose
   edges are blocks intersecting in [lo,hi] points. */
{
    int mb,q,qq,j,t,m,ie,deg;
    long bi;
    setword w;
    static SG_DECL(sg);
    int *ee,*dd;
    size_t *vv;
    char *s6;
    setword block[MAXBLOCKS];

    mb = des[0];

    SG_ALLOC(sg,b,b*b,"proc2des");
    SG_VDE(&sg,vv,dd,ee);
    sg.nv = b;
    // sg.nde = 2*ne;

    j = 0;
    for (q = 1; q <= mb; ++q)
    {
        m = MULT(des[q]);
        w = BLOCK(des[q]);
        for (t = 0; t < m; ++t) block[j++] = w;
    }

    ie = 0;
    for (q = 0; q < b; ++q)
    {
        vv[q] = ie;
        deg = 0;
        w = block[q];
        for (qq = 0; qq < b; ++qq)
        if (qq != q)
        {
            bi = POPCOUNT(w & block[qq]);
            if (bi >= lo && bi <= hi)
            {
                ++deg;
                ee[ie++] = qq;
            }
        }
        dd[q] = deg;
    }

    sg.nde = 2*ie;

    ++nout;
    
    s6 = sgtos6(&sg);

    if (gzipout)
    {
#ifdef ZLIB
        gzwrite(outfilegz,s6,s6len);
#endif
    }
    else
        fwrite(s6,1,s6len,outfile);
}

/*******************************************************************/

static int
readdecimal(char **pos)
/* Find positive decimal integer possibly preceded by spaces.
   Leave the pointer positioned at the following character. */
{
    unsigned long long longval;
    int val;
    char *p;

    p = *pos;
    while (*p == ' ') ++p;
    if (!isdigit(*p)) gt_abort(">E proc2des: missing integer\n");

    longval = 0;
    while (isdigit(*p)) 
    {
        longval = 10*longval + (*p - '0');
        ++p;
    }

    val = (int)longval;
    if (val != longval) gt_abort(">E proc2des: integer too big\n");

    *pos = p;
    return val;
}

static setword
readhex(char **pos)
/* Find hexadecimal integer possibly preceded by spaces.
   Leave the pointer positioned at the following character. */
{
    int count,maxdig;
    setword val,digval;
    char *p;

    maxdig = 2*sizeof(setword);

    p = *pos;
    while (*p == ' ') ++p;
    if (!isxdigit(*p)) gt_abort(">E proc2des: missing hex integer\n");

    val = 0;
    count = 0;
    while (isxdigit(*p))
    {
        digval = (isdigit(*p) ? *p - '0' :
                  isupper(*p) ? *p - 'A' + 10 : *p - 'a' + 10);
        ++p;
        ++count;
        if (count > maxdig)
            gt_abort(">E proc2des: hex integer too long\n"); 
        val = (val << 4 ) | digval;
    }

    *pos = p;
    return val;
}

/*******************************************************************/

static void
extractdesign(design *des, int *vv, int *bb)
/* Extract a design from line[]. Abort on severe errors.  */
{
    int v,b;
    int i,mb,q,sh;
    setword w,wlast,msk;
    char *p;

    p = line;
    v = readdecimal(&p);
    if (v > MAXPOINTS) gt_abort(">E too many points\n");
    b = readdecimal(&p);
    if (b > MAXBLOCKS) gt_abort(">E too many blocks\n");

    *vv = v;
    *bb = b;

    if (b == 0)
    {
        des[0] = 0;
        if (*p != '\n') hasextra = TRUE;
        return;
    }

    sh = WORDSIZE - v;
    msk = ALLMASK(sh);

    q = 0;
    for (i = 0; i < b; ++i)
    {
        w = readhex(&p);
        if ((w & msk)) gt_abort(">E input block has invalid bits\n");
        if (i > 0 && w == wlast) des[q] += 1;
        else                     des[++q] = (w << sh) | 1;
        wlast = w;
    }
    des[0] = mb = q;

    if (*p != '\n') hasextra = TRUE;
    if (v == 0) return;

    shell_des(des+1,mb);

    for (q = 2; q <= mb; ++q)
        if (BLOCK(des[q]) == BLOCK(des[q-1]))
            gt_abort(">E equal input blocks must be together\n");
}

static void
findparams(design *des, int v, int b, paramblock *param)
/* Check if it is a 2-design and fill in all the parameters.
   If it is not a 2-design, only fields v,b,simple,valid are set. */
{
    int mb,q,k,r,lambda,i,j,tot;
    setword msk;

    mb = des[0];

    param->v = v;
    param->b = b;
    param->r = param->k = param->lambda = 0;
    param->simple = (b == mb);
    param->valid = TRUE;

    if (mb == 0) return;
  
    k = BLOCKLEN(des[1]);
    for (q = 1; q <= mb; ++q)
        if (BLOCKLEN(des[q]) != k)
        {
            param->valid = FALSE;
            return;
        }

    if (b*k % v != 0)
    {
        param->valid = FALSE;
        return;
    }
    else
    {
        r = b*k/v;

        for (i = 0; i < v; ++i)
        {
            msk = bit[i];
            tot = 0;
            for (q = 1; q <= mb; ++q)
                if ((des[q]&msk)) tot += MULT(des[q]);
            if (tot != r)
            {
                param->valid = FALSE;
                return;
            }
        }
    }

    if (v > 1)
    {
        if (r*(k-1) % (v-1) != 0)
        {
            param->valid = FALSE;
            return;
        }
        lambda = r*(k-1)/(v-1);
        for (j = 1; j < v; ++j)
        for (i = 0; i < j; ++i)
        {
            msk = bit[i] | bit[j];
            tot = 0;
            for (q = 1; q <= mb; ++q)
                if ((des[q]&msk) == msk) tot += MULT(des[q]);
            if (tot != lambda)
            {
                param->valid = FALSE;
                return;
            }
        }
    }

    param->k = k;
    param->r = r;
    param->lambda = lambda;
}

/**************************************************************************/

static void
addgroup(unsigned long long groupsz, boolean issimple, boolean istransitive)
/* Add group info to table */ 
{
    int i,j;

    for (i = 0; i < numgroups && groupcount[i].groupsize < groupsz; ++i) { }

    if (i < numgroups && groupcount[i].groupsize == groupsz)
    {
        ++groupcount[i].totalnum;
        if (issimple) ++groupcount[i].simplenum;
        if (istransitive) 
        {
            ++groupcount[i].trans;
            if (issimple) ++groupcount[i].simpletrans;
        }
        return;
    }

    if (numgroups == MAXGROUPS)
        gt_abort(">E proc2des : increase MAXGROUPS\n");

    for (j = numgroups; j > i; --j)
        groupcount[j] = groupcount[j-1];
    groupcount[i].groupsize = groupsz;
    groupcount[i].totalnum = 1;
    groupcount[i].simplenum = (issimple ? 1 : 0);
    if (istransitive)
    {
        groupcount[i].trans = 1;
        groupcount[i].simpletrans = (issimple ? 1 : 0);
    }
    else
        groupcount[i].trans = groupcount[i].simpletrans = 0;
    ++numgroups;
}

/**************************************************************************/

static void
addparam(paramblock *param)
/* Add parameter info to table */ 
{
    int i;

    for (i = 0; i < numparams; ++i)
        if (paramcount[i].v == param->v && paramcount[i].b == param->b
             && (paramcount[i].lambda == param->lambda ||
                   (!paramcount[i].valid && !param->valid)))
            break;

    if (i < numparams)
    {
        ++paramcount[i].totalnum;
        if (param->simple) ++paramcount[i].simplenum;
        return;
    }

    if (numparams == MAXPARAMS)
        gt_abort(">E proc2des : increase MAXPARAMS\n");

    paramcount[numparams].v = param->v;
    paramcount[numparams].b = param->b;
    paramcount[numparams].r = param->r;
    paramcount[numparams].k = param->k;
    paramcount[numparams].lambda = param->lambda;
    paramcount[numparams].valid = param->valid;
    paramcount[numparams].totalnum = 1;
    paramcount[numparams].simplenum = (param->simple ? 1 : 0);
    ++numparams;
}

/*******************************************************************/

static void
nullinv(design *des, int s, int *inv)
/* Calculate an invariant; this one does nothing */
{
    int i;

    for (i = 0; i < s; ++i) inv[i] = 0;
}

static void
tripinv(design *des, int s, int *inv)
/* Calculate an invariant based on triple coverage */
{
    int mb,i,j,k,q,count;
    setword mask;

    mb = des[0];
    for (i = 0; i < s; ++i) inv[i] = 0;

    for (k = 2; k < s; ++k)
    for (j = 1; j < k; ++j)
    for (i = 0; i < j; ++i)
    {
        mask = bit[i] | bit[j] | bit[k];
        count = 0;
        for (q = 1; q <= mb; ++q)
            if ((des[q]&mask) == mask) ++count;
        count = (17*count) ^ 5671;
        inv[i] += count;
        inv[j] += count;
        inv[k] += count;
    }
}

static void
quadinv(design *des, int s, int *inv)
/* Calculate an invariant based on quadruple coverage */
{
    int mb,i,j,k,l,q,count;
    setword mask;

    mb = des[0];
    for (i = 0; i < s; ++i) inv[i] = 0;

    for (l = 3; l < s; ++l)
    for (k = 2; k < l; ++k)
    for (j = 1; j < k; ++j)
    for (i = 0; i < j; ++i)
    {
        mask = bit[i] | bit[j] | bit[k] | bit[l];
        count = 0;
        for (q = 1; q <= mb; ++q)
            if ((des[q]&mask) == mask) ++count;
        count ^= 5671;
        for (q = 1; q <= mb; ++q)
            if ((des[q]&mask) == mask)
            {
                inv[i] += count;
                inv[j] += count;
                inv[k] += count;
                inv[l] += count;
            }
    }
}

/*******************************************************************/

static void
makegroupsize(int *lab, int *ptn, int level, int *orbits, statsblk *stats,
              int tv, int index, int tcellsize, int numcells, int cc, int n)
/* Accumulate group size; values >= 2^64 are set to 0; called by nauty */
{
    unsigned long long oldgroupsize;

    oldgroupsize = groupsize;
    groupsize = oldgroupsize * (unsigned long long)index;
    if (groupsize / (unsigned long long)index != oldgroupsize)
        groupsize = 0;
}

/*******************************************************************/

static void
findgroup(design *des, int s, design *candes,
                        int *numorbits, unsigned long long *automs)
/* Compute the group size (0 if more than 2^64-1), the number of
   orbits, and a canonical labelling. */
{
    int lab[MAXPOINTS],ptn[MAXPOINTS],orbits[MAXPOINTS];
    int invalue[MAXPOINTS];
    statsblk stats;
    static DEFAULTOPTIONS_MDES(options);

    INVAR(des,s,invalue);
    setlabptn(invalue,lab,ptn,s);

    options.getcanon = TRUE;
    options.userlevelproc = makegroupsize;
    options.defaultptn = FALSE;
 
    groupsize = 1;
    naumdesign(des,lab,ptn,NULL,orbits,&options,&stats,s,candes);
 
    *numorbits = stats.numorbits;
    *automs = groupsize;
}

/*******************************************************************/

static boolean
isgzfile(char *fname)
/* Check if fname ends in .gz */
{
    int i;

    if (fname == NULL) return FALSE;

    for (i = 0; fname[i] != '\0' ; ++i)
    {
        if (fname[i] == '.' && fname[i+1] == 'g'
                            && fname[i+2] == 'z' && fname[i+3] == '\0')
            return TRUE;
    }

    return FALSE;
}

static void
openinfile(char *fname, long initpos, boolean fswitch)
/* Open input file for reading. If its name indicates it is gzipped,
   set infilegz and gzipin. Otherwise set infile and unset gzipin.
   If initpos > 1, skip to that line. */
{
    long pos;
    off_t off0,off1,off2;

    if (fname == NULL || (fname[0] == '-' && fname[1] == '\0'))
    {
        infile = stdin;
        gzipin = FALSE;
        return;
    } 

    if (isgzfile(fname))
    {
#ifdef ZLIB
        gzipin = TRUE;
        if ((infilegz = gzopen(fname,"r")) == NULL)
            gt_abort(">E unable to open input file\n");

        for (pos = 1; pos < initpos; ++pos)
            if (gzgets(infilegz,line,LINELEN-2) == NULL)
                gt_abort(">E Initial position is past EOF\n");
#else
        gt_abort(">E proc2des: gzipped input is only available"
                 " if proc2des is compiled with -DZLIB\n");
#endif
    }
    else
    {
        gzipin = FALSE;
        if ((infile = fopen(fname,"r")) == NULL)
            gt_abort(">E unable to open input file\n");

        if (initpos == 1) return;

        if (fswitch)
        {
            off0 = ftello(infile);
            if (fgets(line,LINELEN-2,infile) == NULL)
                gt_abort(">E Initial position is past EOF\n");
            if (initpos == 2) return;
            off1 = ftello(infile);

            if (off0 < 0 || off1 < 0)  /* Unseekable stream, eg. tty or pipe */
            {        
                for (pos = 1; pos < initpos; ++pos)
                    if (fgets(line,LINELEN-2,infile) == NULL)
                        gt_abort(">E Initial position is past EOF\n");
            }
            else
            {
                off2 = off0 + (initpos-1)*(off1-off0);
                if (fseeko(infile,off2,SEEK_SET) < 0)
                    gt_abort(">E Initial seek failed; try without -f\n");
            }
        }
        else
        {
            for (pos = 1; pos < initpos; ++pos)
                if (fgets(line,LINELEN-2,infile) == NULL)
                    gt_abort(">E Initial position is past EOF\n");
        }
    }
}

static void
openoutfile(char *fname, boolean gzip)
/* Open output file, as gzipped stream if either gzip or fname
   ends in ".gz". Either outfile or outfilegz is defined and
   gzipout is set or unset. */
{
    if (fname == NULL || (fname[0] == '-' && fname[1] == '\0'))
    {
        if (gzip)
        {
#ifdef ZLIB
            if (isatty(fileno(stdout)))
                gt_abort(
       ">E proc2des: probably you don't want gzip output to your terminal\n");
            outfilegz = gzdopen(fileno(stdout),"w");
            if (outfilegz == NULL)
                gt_abort(">E gen2des: failed to open gzip stream to stdout\n");
            gzipout = TRUE;
#else
            gt_abort(
             ">E proc2des: gzipped output requires compilation with -DZLIB\n");
#endif
        }
        else
        {
            outfile = stdout;
            gzipout = FALSE;
        }
        return;
    }

    if (gzip || isgzfile(fname))
    {
#ifdef ZLIB
        if ((outfilegz = gzopen(fname,"w")) == NULL)
            gt_abort(">E gen2des: failed to open output file\n");
        gzipout = TRUE;
#else
        gt_abort(
          ">E proc2des: gzipped output requires compilation with -DZLIB\n");
#endif
    }
    else
    {
        if ((outfile = fopen(fname,"w")) == NULL)
            gt_abort(">E gen2des: failed to open output file\n");
        gzipout = FALSE;
    }
}

static boolean
readnextline()
/* Read a line into line[], return success */
{
    if (gzipin)
    {
#if ZLIB
        if (gzgets(infilegz,line,LINELEN-2) == NULL)
            return FALSE;
#endif
    }
    else
    {
        if (fgets(line,LINELEN-2,infile) == NULL)
            return FALSE;
    }

    return TRUE;
}

/*******************************************************************/

#ifdef SHA
static void
writehash(design *des)
/* Write a 20-character hash code to outfile/outfilegz */
{
    unsigned char raw[SHA_DIGEST_LENGTH+1];
    char cooked[SHA_DIGEST_LENGTH+2];
    int mb,i;

    mb = des[0];
    SHA1((unsigned char*)des,(mb+1)*sizeof(design),raw);
    for (i = 0; i < SHA_DIGEST_LENGTH; ++i)
        cooked[i] = 33 + (raw[i]%94);

    cooked[SHA_DIGEST_LENGTH] = '\n';
    cooked[SHA_DIGEST_LENGTH+1] = '\0';

    if (gzipout)
    {
#ifdef ZLIB
        gzwrite(outfilegz,cooked,SHA_DIGEST_LENGTH+1);
#endif
    }
    else
        fwrite(cooked,1,SHA_DIGEST_LENGTH+1,outfile);

    ++nout;
}
#endif

/*******************************************************************/

static void
docomplement(design *des, int v)
/* Replace des by its block-wise complement */
{
    int mb,q;
    setword msk;

    mb = des[0];
    msk = ALLMASK(v);

    for (q = 1; q <= mb; ++q) des[q] ^= msk;
    shell_des(des+1,mb);
}

static void
makedual(design *des, int v, int b)
/* Replace des by its dual (transpose). This is only allowed
   if des is simple and v=b. */
{
    int mb,q,i;
    design work[MAXPOINTS+2];
    setword bqm,w;

    mb = des[0];

    for (i = 0; i < mb; ++i) work[i] = 0;

    for (q = 1; q <= mb; ++q)
    {
        bqm = bit[q-1];
        w = BLOCK(des[q]);
        while (w)
        {
            TAKEBIT(i,w);
            work[i] |= bqm;
        }
    }

    shell_des(work,mb);
    for (i = 0; i < mb; ++i) des[i+1] = work[i] | 1;
}

/*******************************************************************/

static void
makeresidual(design *des, int v, int nb, design *res, int *newv)
/* Make residual of symmetric design using block nb */
{
    int mb,i,q,qr,j;
    setword block,notblock,w,newblock;
    setword p[MAXPOINTS];

    mb = des[0];
    block = BLOCK(des[nb+1]);
    notblock = ALLMASK(v) ^ block;

    j = 0;
    w = notblock;
    while (w)
    {
        TAKEBIT(i,w);
        p[i] = bit[j++];
    }
     

    qr = 0;
    for (q = 1; q <= mb; ++q)
    if (q != nb+1)
    {
        w = des[q] & notblock;
        newblock = 0;
        while (w)
        {
            TAKEBIT(i,w);
            newblock |= p[i];
        }
        res[++qr] = newblock;
    }

    shell_des(res+1,qr);

    qr = 0;
    for (q = 1; q < mb; ++q)
    {
        if (qr == 0 || res[q] != BLOCK(res[qr]))
            res[++qr] = res[q] | 1;
        else
            ++res[qr];
    }

    res[0] = qr;
    *newv = v - POPCOUNT(block);
}

static void
makederived(design *des, int v, int nb, design *der, int *newv)
/* Make derived design of symmetric design using block nb.
   nb is 0..v-1. */
{
    int mb,i,q,qr,j;
    setword block,w,newblock;
    setword p[MAXPOINTS];

    mb = des[0];
    block = BLOCK(des[nb+1]);

    j = 0;
    w = block;
    while (w)
    {
        TAKEBIT(i,w);
        p[i] = bit[j++];
    }
     
    qr = 0;
    for (q = 1; q <= mb; ++q)
    if (q != nb+1)
    {
        w = des[q] & block;
        newblock = 0;
        while (w)
        {
            TAKEBIT(i,w);
            newblock |= p[i];
        }
        der[++qr] = newblock;
    }

    shell_des(der+1,qr);

    qr = 0;
    for (q = 1; q < mb; ++q)
    {
        if (qr == 0 || der[q] != BLOCK(der[qr]))
            der[++qr] = der[q] | 1;
        else
            ++der[qr];
    }

    der[0] = qr;
    *newv = POPCOUNT(block);
}

/*******************************************************************/

int
main(int argc, char *argv[])
{
    boolean badargs,zswitch,getcanon,resid,dual,matrix,Matrix,list;
    boolean quiet,verbose,Verbose,fswitch,pswitch,oswitch,aswitch;
    boolean needgroup,chosen,hash,complement,derived,bipartite;
    boolean bigraph;
    int argnum,i,j,numorbits,v,b,vv,bb;
    long pos,pos1,pos2,aut1,aut2,orb1,orb2,bilo,bihi;
    int nb,numcases;
    unsigned long long automs;
    char *arg,sw;
    char *infilename,*outfilename;
    double t0,t1;
    char msg[101];
    design des[MAXBLOCKS+1],candes[MAXBLOCKS+1];
    design newdes[MAXBLOCKS+1],*thisdes,*qdes;
    paramblock param;
    long long nin;
    FILE *vfile;

    HELP; PUTVERSION;

    naumdesign_check(WORDSIZE,1,WORDSIZE-MULTBITS,NAUTYVERSIONID);

    getcanon = simple = quiet = uswitch = verbose = Verbose = pswitch = FALSE;
    fswitch = oswitch = zswitch = gzipin = gzipout = hasextra = FALSE;
    resid = dual = complement = derived = bipartite = FALSE;
    matrix = Matrix = list = aswitch = hash = bigraph = FALSE;
    infilename = outfilename = NULL;

    argnum = 0;
    badargs = FALSE;
    for (j = 1; !badargs && j < argc; ++j)
    {
        arg = argv[j];
        if (arg[0] == '-' && arg[1] != '\0')
        {
            ++arg;
            while (*arg != '\0')
            {
                sw = *arg++;
                     SWBOOLEAN('s',simple)
                else SWBOOLEAN('q',quiet)
                else SWBOOLEAN('u',uswitch)
                else SWBOOLEAN('f',fswitch)
                else SWBOOLEAN('l',getcanon)
                else SWBOOLEAN('R',resid)
                else SWBOOLEAN('D',derived)
                else SWBOOLEAN('c',complement)
                else SWBOOLEAN('d',dual)
                else SWBOOLEAN('m',matrix)
                else SWBOOLEAN('M',Matrix)
                else SWBOOLEAN('w',list)
                else SWBOOLEAN('B',bipartite)
                else SWRANGE('b',":-",bigraph,bilo,bihi,"proc2des -b")
                else SWRANGE('p',":-",pswitch,pos1,pos2,"proc2des -p")
                else SWRANGE('a',":-",aswitch,aut1,aut2,"proc2des -a")
                else SWRANGE('o',":-",oswitch,orb1,orb2,"proc2des -o")
                else SWBOOLEAN('v',verbose)
                else SWBOOLEAN('V',Verbose)
#ifdef ZLIB
                else SWBOOLEAN('z',zswitch)
#endif
#ifdef SHA
                else SWBOOLEAN('h',hash)
#endif
                else badargs = TRUE;
            }
        }
        else
        {
            ++argnum;
            if (argnum == 1)
                infilename = arg;
            else if (argnum == 2)
                outfilename = arg;
            else
                badargs = TRUE;
        }
    }

    if (badargs)
    {
        fprintf(stderr,">E Usage: %s\n",USAGE);
        GETHELP;
        exit(1);
    }

    if (!pswitch)
    {
        pos1 = 1;
        pos2 = -1;
    }
    else
    {
        if (pos1 < 1) pos1 = 1;
        if (pos2 == 0 || pos2 == NOLIMIT) pos2 = -1;
        if (pos2 > 0 && pos2 < pos1)
            gt_abort(">E proc2des -p: empty range\n");
    }

    if (aswitch && aut1 < 1) aut1 = 1;
    if (oswitch && orb1 < 1) orb1 = 1;

    if ((uswitch==TRUE) + (matrix==TRUE) + (Matrix==TRUE)
         + (bipartite==TRUE) + (list==TRUE) + (hash==TRUE) > 1)
        gt_abort(">E proc2des: -u, -B, -m, -M, -w, -h are incompatible\n");
    if (derived && resid)
        gt_abort(">E proc2des: -D and -R are incompatible\n");

    if (Verbose) verbose = TRUE;

    if (uswitch) getcanon = FALSE;
    needgroup = (Verbose || verbose || getcanon
                         || aswitch || oswitch || hash);

    openinfile(infilename,pos1,fswitch);
    if (!uswitch) openoutfile(outfilename,zswitch);

    if (infilename == NULL) infilename = "-";
    if (outfilename == NULL) outfilename = "-";

    if (!quiet)
    {
        msg[0] = '\0';
        CATMSG0(">A proc2des");
        if (simple) CATMSG0(" -s");
        if (resid || dual) CATMSG3(" -%s%s%s",
            (resid?"R":""),(dual?"d":""),(derived?"D":""));
        if (uswitch || matrix || Matrix || list || bipartite)
            CATMSG5(" -%s%s%s%s%s",(uswitch?"u":""),(bipartite?"B":""),
                (matrix?"m":""),(Matrix?"M":""),(list?"w":""));
        if (aswitch)
        {
            if (aut2 == NOLIMIT) CATMSG1(" -a%ld:",aut1);
            else                 CATMSG2(" -a%ld:%ld",aut1,aut2);
        }
        if (oswitch)
        {
            if (orb2 == NOLIMIT) CATMSG1(" -o%ld:",orb1);
            else                 CATMSG2(" -o%ld:%ld",orb1,orb2);
        }
        if (getcanon) CATMSG0(" -l");
        if (pswitch) CATMSG2(" -p%ld:%ld",pos1,pos2);
        fprintf(stderr,"%s %s %s\n",msg,infilename,outfilename);
    }

    numgroups = 0;
    nin = nout = nsimple = 0;

    t0 = CPUTIME;

    for (pos = pos1; ; ++pos)
    {
        if (!readnextline()) break;
        ++nin;

        extractdesign(des,&v,&b);

        if ((dual || resid || derived) && (v != b || b != des[0]))
            gt_abort(">E proc2des: -d, -D, -R are only allowed"
                     " for simple designs with v=b\n");

        if (complement) docomplement(des,v);
        if (dual) makedual(des,v,b);

        if (resid || derived) numcases = b;
        else                  numcases = 1;

        for (nb = 0; nb < numcases; ++nb)
        {
            if (resid)
            {
                makeresidual(des,v,nb,newdes,&vv);
                thisdes = newdes;
                bb = b - 1;
            }
            else if (derived)
            {
                makederived(des,v,nb,newdes,&vv);
                thisdes = newdes;
                bb = b - 1;
            }
            else
            {
                thisdes = des;
                vv = v;
                bb = b;
            }

            findparams(thisdes,vv,bb,&param);

            chosen = TRUE;
            if (simple && !param.simple) chosen = FALSE;
            if (chosen && needgroup)
            {
                findgroup(thisdes,vv,candes,&numorbits,&automs);
                if (oswitch && (numorbits < orb1 || numorbits > orb2))
                chosen = FALSE;
                if (aswitch)
                {
                    if (automs < aut1 && automs != 0) chosen = FALSE;
                    if (aut1 != NOLIMIT && (automs == 0 || automs > aut2))
                        chosen = FALSE;
                }
                if (chosen || Verbose)
                    addgroup(automs,param.simple,(numorbits==1));
            }
            if (chosen)
            {
                qdes = (getcanon ? candes : thisdes);
                addparam(&param);
                if (matrix)
                    writematrix(pos,qdes,vv,bb,TRUE);
                else if (Matrix)
                    writematrix(pos,qdes,vv,bb,FALSE);
                else if (list)
                    writeblocks(pos,qdes,vv,bb);
                else if (bipartite)
                    writeincidence(qdes,vv,bb);
                else if (bigraph)
                    writebigraph(qdes,vv,bb,bilo,bihi);
#ifdef SHA
                else if (hash)
                    writehash(candes);
#endif
                else
                    writedes(qdes,&param);
            }
        }
        if (pos == pos2) break;
    }

    t1 = CPUTIME;

#ifdef ZLIB
    if (gzipout)
    {
        if (gzclose(outfilegz) != Z_OK)
            gt_abort(">E proc2des: Error on closing gzip stream\n");
    }
#endif

    if (verbose)
    {
        vfile = (uswitch ? stdout : stderr);
        fprintf(vfile,">G   |Aut(D)|    designs     simple  trans simtrans\n");
        for (i = 0; i < numgroups; ++i)
            fprintf(vfile,">G %10llu %10lld %10lld %5lld %5lld\n",
                groupcount[i].groupsize,
                groupcount[i].totalnum,groupcount[i].simplenum,
                groupcount[i].trans,groupcount[i].simpletrans);
    }

    for (i = 0; i < numparams; ++i)
    {
        if (paramcount[i].valid)
            fprintf(stderr,">C v=%d b=%d r=%d k=%d"
                           " lambda=%d : %lld (%lld simple)\n",
                paramcount[i].v,paramcount[i].b,paramcount[i].r,
                paramcount[i].k,paramcount[i].lambda,
                paramcount[i].totalnum,paramcount[i].simplenum);
        else
            fprintf(stderr,">C v=%d b=%d not a 2-design :"
                           " %lld (%lld simple)\n",
                paramcount[i].v,paramcount[i].b,
                paramcount[i].totalnum,paramcount[i].simplenum);
    }

    if (hasextra)
        fprintf(stderr,">W proc2des : some input lines have extra data\n");

    if (uswitch)
        fprintf(stderr,">Z %lld designs read from %s;"
                " %lld selected; %.2f sec\n",
                nin,infilename,nout,t1-t0);
    else
        fprintf(stderr,">Z %lld designs read from %s;"
                   " %lld written to %s; %.2f sec\n",
                nin,infilename,nout,outfilename,t1-t0);

    exit(0);
}
